from autoarray.plot.mat_wrap import mat_plot as mp
from autoarray.plot.plotters import inversion_plotters
from autogalaxy.plot.mat_wrap import lensing_mat_plot, lensing_include, lensing_visuals
from autogalaxy.plot.plotters import plane_plotters
from autoarray.plot.plotters import fit_interferometer_plotters
from autogalaxy.fit import fit as f


class FitInterferometerPlotter(
    fit_interferometer_plotters.AbstractFitInterferometerPlotter
):
    def __init__(
        self,
        fit: f.FitInterferometer,
        mat_plot_1d: lensing_mat_plot.MatPlot1D = lensing_mat_plot.MatPlot1D(),
        visuals_1d: lensing_visuals.Visuals1D = lensing_visuals.Visuals1D(),
        include_1d: lensing_include.Include1D = lensing_include.Include1D(),
        mat_plot_2d: lensing_mat_plot.MatPlot2D = lensing_mat_plot.MatPlot2D(),
        visuals_2d: lensing_visuals.Visuals2D = lensing_visuals.Visuals2D(),
        include_2d: lensing_include.Include2D = lensing_include.Include2D(),
    ):

        super().__init__(
            fit=fit,
            mat_plot_1d=mat_plot_1d,
            include_1d=include_1d,
            visuals_1d=visuals_1d,
            mat_plot_2d=mat_plot_2d,
            include_2d=include_2d,
            visuals_2d=visuals_2d,
        )

    @property
    def plane(self):
        return self.fit.plane

    @property
    def visuals_with_include_2d(self):
        visuals_2d = super(FitInterferometerPlotter, self).visuals_with_include_2d
        return visuals_2d + visuals_2d.__class__()

    def plane_plotter_from(self, plane):
        return plane_plotters.PlanePlotter(
            plane=plane,
            grid=self.fit.masked_interferometer.grid,
            mat_plot_2d=self.mat_plot_2d,
            visuals_2d=self.visuals_with_include_2d,
            include_2d=self.include_2d,
        )

    @property
    def inversion_plotter(self):
        return inversion_plotters.InversionPlotter(
            inversion=self.fit.inversion,
            mat_plot_2d=self.mat_plot_2d,
            visuals_2d=self.visuals_with_include_2d,
            include_2d=self.include_2d,
        )

    def subplot_fit_real_space(self):

        if self.fit.inversion is None:

            plane_plotter = self.plane_plotter_from(plane=self.plane)

            plane_plotter.subplot(
                image=True, plane_image=True, auto_filename="subplot_fit_real_space"
            )

        elif self.fit.inversion is not None:

            self.inversion_plotter.subplot(
                reconstructed_image=True,
                reconstruction=True,
                auto_filename="subplot_fit_real_space",
            )



from os import path

import pytest

import autofit as af
import autogalaxy as ag
from autogalaxy.mock import mock

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(name="path")
def make_path():
    return path.join("{}".format(path.dirname(path.realpath(__file__))), "files")


@pytest.fixture(name="samples")
def make_samples():
    galaxy_0 = ag.Galaxy(redshift=0.5, light=ag.lp.EllipticalSersic(centre=(0.0, 1.0)))
    galaxy_1 = ag.Galaxy(redshift=1.0, light=ag.lp.EllipticalSersic())

    plane = ag.Plane(galaxies=[galaxy_0, galaxy_1])

    return mock.MockSamples(max_log_likelihood_instance=plane)


def test__dataset_generator_from_aggregator(imaging_7x7, mask_7x7, samples):
    phase_imaging_7x7 = ag.PhaseImaging(
        galaxies=dict(
            galaxy=ag.GalaxyModel(redshift=0.5, light=ag.lp.EllipticalSersic),
            source=ag.GalaxyModel(redshift=1.0, light=ag.lp.EllipticalSersic),
        ),
        search=mock.MockSearch(samples=samples, name="test_phase_aggregator"),
    )

    imaging_7x7.positions = ag.Grid2DIrregularGrouped([[1.0, 1.0], [2.0, 2.0]])

    phase_imaging_7x7.run(
        dataset=imaging_7x7, mask=mask_7x7, results=mock.MockResults(samples=samples)
    )

    agg = af.Aggregator(directory=phase_imaging_7x7.paths.output_path)

    dataset = list(agg.values("dataset"))

    print(dataset)


def test__plane_generator_from_aggregator(imaging_7x7, mask_7x7, samples):
    phase_imaging_7x7 = ag.PhaseImaging(
        galaxies=dict(
            galaxy=ag.GalaxyModel(redshift=0.5, light=ag.lp.EllipticalSersic),
            source=ag.GalaxyModel(redshift=1.0, light=ag.lp.EllipticalSersic),
        ),
        search=mock.MockSearch(samples=samples, name="test_phase_aggregator"),
    )

    phase_imaging_7x7.run(
        dataset=imaging_7x7, mask=mask_7x7, results=mock.MockResults(samples=samples)
    )

    agg = af.Aggregator(directory=phase_imaging_7x7.paths.output_path)

    plane_gen = ag.agg.Plane(aggregator=agg)

    for plane in plane_gen:

        assert plane.galaxies[0].redshift == 0.5
        assert plane.galaxies[0].light.centre == (0.0, 1.0)
        assert plane.galaxies[1].redshift == 1.0


def test__masked_imaging_generator_from_aggregator(imaging_7x7, mask_7x7, samples):
    phase_imaging_7x7 = ag.PhaseImaging(
        galaxies=dict(
            galaxy=ag.GalaxyModel(redshift=0.5, light=ag.lp.EllipticalSersic),
            source=ag.GalaxyModel(redshift=1.0, light=ag.lp.EllipticalSersic),
        ),
        settings=ag.SettingsPhaseImaging(
            settings_masked_imaging=ag.SettingsMaskedImaging(
                grid_class=ag.Grid2DIterate,
                grid_inversion_class=ag.Grid2DIterate,
                fractional_accuracy=0.5,
                sub_steps=[2],
            )
        ),
        search=mock.MockSearch(samples=samples, name="test_phase_aggregator"),
    )

    phase_imaging_7x7.run(
        dataset=imaging_7x7, mask=mask_7x7, results=mock.MockResults(samples=samples)
    )

    agg = af.Aggregator(directory=phase_imaging_7x7.paths.output_path)

    masked_imaging_gen = ag.agg.MaskedImaging(aggregator=agg)

    for masked_imaging in masked_imaging_gen:
        assert (masked_imaging.imaging.image == imaging_7x7.image).all()
        assert isinstance(masked_imaging.grid, ag.Grid2DIterate)
        assert isinstance(masked_imaging.grid_inversion, ag.Grid2DIterate)
        assert masked_imaging.grid.sub_steps == [2]
        assert masked_imaging.grid.fractional_accuracy == 0.5


def test__fit_imaging_generator_from_aggregator(imaging_7x7, mask_7x7, samples):
    phase_imaging_7x7 = ag.PhaseImaging(
        galaxies=dict(
            galaxy=ag.GalaxyModel(redshift=0.5, light=ag.lp.EllipticalSersic),
            source=ag.GalaxyModel(redshift=1.0, light=ag.lp.EllipticalSersic),
        ),
        search=mock.MockSearch(samples=samples, name="test_phase_aggregator"),
    )

    phase_imaging_7x7.run(
        dataset=imaging_7x7, mask=mask_7x7, results=mock.MockResults(samples=samples)
    )

    agg = af.Aggregator(directory=phase_imaging_7x7.paths.output_path)

    fit_imaging_gen = ag.agg.FitImaging(aggregator=agg)

    for fit_imaging in fit_imaging_gen:
        assert (fit_imaging.masked_imaging.imaging.image == imaging_7x7.image).all()


def test__masked_interferometer_generator_from_aggregator(
    interferometer_7, visibilities_mask_7, mask_7x7, samples
):
    phase_interferometer_7x7 = ag.PhaseInterferometer(
        galaxies=dict(
            galaxy=ag.GalaxyModel(redshift=0.5, light=ag.lp.EllipticalSersic),
            source=ag.GalaxyModel(redshift=1.0, light=ag.lp.EllipticalSersic),
        ),
        settings=ag.SettingsPhaseInterferometer(
            settings_masked_interferometer=ag.SettingsMaskedInterferometer(
                grid_class=ag.Grid2DIterate,
                grid_inversion_class=ag.Grid2DIterate,
                fractional_accuracy=0.5,
                sub_steps=[2],
                transformer_class=ag.TransformerDFT,
            )
        ),
        search=mock.MockSearch(samples=samples, name="test_phase_aggregator"),
        real_space_mask=mask_7x7,
    )

    phase_interferometer_7x7.run(
        dataset=interferometer_7,
        mask=visibilities_mask_7,
        results=mock.MockResults(samples=samples),
    )

    agg = af.Aggregator(directory=phase_interferometer_7x7.paths.output_path)

    masked_interferometer_gen = ag.agg.MaskedInterferometer(aggregator=agg)

    for masked_interferometer in masked_interferometer_gen:
        assert (
            masked_interferometer.interferometer.visibilities
            == interferometer_7.visibilities
        ).all()
        assert (masked_interferometer.real_space_mask == mask_7x7).all()
        assert isinstance(masked_interferometer.grid, ag.Grid2DIterate)
        assert isinstance(masked_interferometer.grid_inversion, ag.Grid2DIterate)
        assert masked_interferometer.grid.sub_steps == [2]
        assert masked_interferometer.grid.fractional_accuracy == 0.5
        assert isinstance(masked_interferometer.transformer, ag.TransformerDFT)


def test__fit_interferometer_generator_from_aggregator(
    interferometer_7, visibilities_mask_7, mask_7x7, samples
):
    phase_interferometer_7x7 = ag.PhaseInterferometer(
        galaxies=dict(
            galaxy=ag.GalaxyModel(redshift=0.5, light=ag.lp.EllipticalSersic),
            source=ag.GalaxyModel(redshift=1.0, light=ag.lp.EllipticalSersic),
        ),
        search=mock.MockSearch(samples=samples, name="test_phase_aggregator"),
        real_space_mask=mask_7x7,
    )

    phase_interferometer_7x7.run(
        dataset=interferometer_7,
        mask=visibilities_mask_7,
        results=mock.MockResults(samples=samples),
    )

    agg = af.Aggregator(directory=phase_interferometer_7x7.paths.output_path)

    fit_interferometer_gen = ag.agg.FitInterferometer(aggregator=agg)

    for fit_interferometer in fit_interferometer_gen:
        assert (
            fit_interferometer.masked_interferometer.interferometer.visibilities
            == interferometer_7.visibilities
        ).all()
        assert (
            fit_interferometer.masked_interferometer.real_space_mask == mask_7x7
        ).all()



from astropy import cosmology as cosmo
from autoconf import conf
import autofit as af
from autoarray.inversion import pixelizations as pix
from autoarray.inversion import regularization as reg
from autogalaxy.pipeline.phase import abstract
from autogalaxy.pipeline.phase import extensions
from autogalaxy.pipeline.phase.dataset.result import Result

import copy
import os
import shutil
from distutils.dir_util import copy_tree


class PhaseDataset(abstract.AbstractPhase):
    galaxies = af.PhaseProperty("galaxies")

    Result = Result

    def __init__(
        self,
        settings,
        search,
        galaxies=None,
        cosmology=cosmo.Planck15,
        use_as_hyper_dataset=False,
    ):
        """

        A phase in an lens pipeline. Uses the set non_linear search to try to fit models and hyper_galaxies
        passed to it.

        Parameters
        ----------
        search: class
            The class of a non_linear search
        """

        super().__init__(
            search=search, settings=settings, galaxies=galaxies, cosmology=cosmology
        )

        self.hyper_name = None
        self.use_as_hyper_dataset = use_as_hyper_dataset
        self.is_hyper_phase = False

    def run(
        self,
        dataset,
        mask,
        results=None,
        info=None,
        pickle_files=None,
        log_likelihood_cap=None,
    ):
        """
        Run this phase.

        Parameters
        ----------
        mask: Mask2D
            The default masks passed in by the pipeline
        results: autofit.tools.pipeline.ResultsCollection
            An object describing the results of the last phase or None if no phase has been executed
        dataset: scaled_array.ScaledSquarePixelArray
            An masked_imaging that has been masked

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising the best fit model and other hyper_galaxies.
        """

        self.model = self.model.populate(results)

        results = results or af.ResultsCollection()

        self.modify_dataset(dataset=dataset, results=results)
        self.modify_settings(dataset=dataset, results=results)
        self.modify_search_paths()

        analysis = self.make_analysis(dataset=dataset, mask=mask, results=results)

        result = self.run_analysis(
            analysis=analysis,
            info=info,
            pickle_files=pickle_files,
            log_likelihood_cap=log_likelihood_cap,
        )

        return self.make_result(result=result, analysis=analysis)

    def make_analysis(self, dataset, mask, results=None):
        """
        Returns an lens object. Also calls the prior passing and masked_imaging modifying functions to allow child
        classes to change the behaviour of the phase.

        Parameters
        ----------
        mask: Mask2D
            The default masks passed in by the pipeline
        dataset: im.Imaging
            An masked_imaging that has been masked
        results: autofit.tools.pipeline.ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens : Analysis
            An lens object that the `NonLinearSearch` calls to determine the fit of a set of values
        """
        raise NotImplementedError()

    def modify_dataset(self, dataset, results):
        pass

    def modify_settings(self, dataset, results):
        pass

    def modify_search_paths(self):
        """
        Modify the output paths of the phase before the non-linear search is run, so that the output path can be
        customized using the tags of the phase.
        """
        if self.hyper_name is None:
            hyper_tag = ""
        else:
            hyper_tag = f"{self.hyper_name}__"

        if not self.has_pixelization:
            self.search.paths.tag = f"{hyper_tag}{self.settings.phase_tag_no_inversion}"
        else:
            self.search.paths.tag = (
                f"{hyper_tag}{self.settings.phase_tag_with_inversion}"
            )

        if self.hyper_name == "hyper":

            rename_hyper_combined = conf.instance["general"]["hyper"][
                "rename_hyper_combined"
            ]

            if rename_hyper_combined:

                output_path_hyper = copy.copy(self.search.paths.output_path)
                output_path_hyper_combined = output_path_hyper.replace(
                    "hyper", "hyper_combined"
                )

                if os.path.exists(output_path_hyper_combined):
                    copy_tree(output_path_hyper_combined, output_path_hyper)
                    if os.path.isfile(f"{output_path_hyper_combined}.zip"):
                        shutil.copyfile(
                            f"{output_path_hyper_combined}.zip",
                            f"{output_path_hyper}.zip",
                        )
                    shutil.rmtree(output_path_hyper_combined)

                if os.path.isfile(f"{output_path_hyper_combined}.zip"):
                    os.remove(f"{output_path_hyper_combined}.zip")

                if os.path.exists(os.path.join(output_path_hyper_combined, "..")):
                    shutil.rmtree(os.path.join(output_path_hyper_combined, ".."))

    def extend_with_hyper_phase(self, setup_hyper, include_hyper_image_sky=True):

        self.use_as_hyper_dataset = True

        if not self.has_pixelization:
            if setup_hyper.hypers_all_off:
                return self
            if setup_hyper.hypers_all_except_image_sky_off:
                if not include_hyper_image_sky:
                    return self

        if self.has_pixelization:
            hyper_search = setup_hyper.hyper_search_with_inversion
        else:
            hyper_search = setup_hyper.hyper_search_no_inversion

        if include_hyper_image_sky:
            hyper_image_sky = setup_hyper.hyper_image_sky
        else:
            hyper_image_sky = None

        return extensions.HyperPhase(
            phase=self,
            hyper_search=hyper_search,
            model_classes=(pix.Pixelization, reg.Regularization),
            hyper_image_sky=hyper_image_sky,
            hyper_background_noise=setup_hyper.hyper_background_noise,
            hyper_galaxy_names=setup_hyper.hyper_galaxy_names,
        )



import autofit as af
import pickle


class Analysis(af.Analysis):
    def __init__(self, cosmology, settings):

        self.cosmology = cosmology
        self.settings = settings

    def save_settings(self, paths: af.Paths):
        with open(f"{paths.pickle_path}/settings.pickle", "wb+") as f:
            pickle.dump(self.settings, f)



from autoarray.structures import grids
from autoarray.plot.mat_wrap import mat_plot as mp
from autoarray.util import plotter_util
from autoarray.plot.plotters import abstract_plotters
from autogalaxy.plot.mat_wrap import lensing_mat_plot, lensing_include, lensing_visuals


class LightProfilePlotter(abstract_plotters.AbstractPlotter):
    def __init__(
        self,
        light_profile,
        grid,
        mat_plot_1d: lensing_mat_plot.MatPlot1D = lensing_mat_plot.MatPlot1D(),
        visuals_1d: lensing_visuals.Visuals1D = lensing_visuals.Visuals1D(),
        include_1d: lensing_include.Include1D = lensing_include.Include1D(),
        mat_plot_2d: lensing_mat_plot.MatPlot2D = lensing_mat_plot.MatPlot2D(),
        visuals_2d: lensing_visuals.Visuals2D = lensing_visuals.Visuals2D(),
        include_2d: lensing_include.Include2D = lensing_include.Include2D(),
    ):

        self.light_profile = light_profile
        self.grid = grid

        super().__init__(
            mat_plot_2d=mat_plot_2d,
            include_2d=include_2d,
            visuals_2d=visuals_2d,
            mat_plot_1d=mat_plot_1d,
            include_1d=include_1d,
            visuals_1d=visuals_1d,
        )

    @property
    def visuals_with_include_2d(self) -> lensing_visuals.Visuals2D:
        """
        Extracts from a `Structure` attributes that can be plotted and return them in a `Visuals` object.

        Only attributes with `True` entries in the `Include` object are extracted for plotting.

        From an `AbstractStructure` the following attributes can be extracted for plotting:

        - origin: the (y,x) origin of the structure's coordinate system.
        - mask: the mask of the structure.
        - border: the border of the structure's mask.

        Parameters
        ----------
        structure : abstract_structure.AbstractStructure
            The structure whose attributes are extracted for plotting.

        Returns
        -------
        vis.Visuals2D
            The collection of attributes that can be plotted by a `Plotter2D` object.
        """

        return self.visuals_2d + self.visuals_2d.__class__(
            origin=self.extract_2d(
                "origin", value=grids.Grid2DIrregular(grid=[self.grid.origin])
            ),
            mask=self.extract_2d("mask", value=self.grid.mask),
            border=self.extract_2d(
                "border", value=self.grid.mask.border_grid_sub_1.slim_binned
            ),
            light_profile_centres=self.extract_2d(
                "light_profile_centres", self.light_profile.light_profile_centres
            ),
        )

    def figures(self, image=False):

        if image:

            self.mat_plot_2d.plot_array(
                array=self.light_profile.image_from_grid(grid=self.grid),
                visuals_2d=self.visuals_with_include_2d,
                auto_labels=mp.AutoLabels(title="Image", filename="image"),
            )

    def figures_1d(self, image=False):

        grid_radii = self.grid.grid_radii_from(centre=self.light_profile.centre)

        if image:

            self.mat_plot_1d.plot_line(
                y=self.light_profile.image_from_grid(grid=grid_radii),
                x=grid_radii[:, 1],
                visuals_1d=self.visuals_1d,
                auto_labels=mp.AutoLabels(
                    title="Image vs Radius",
                    ylabel="Image",
                    xlabel="Radius",
                    legend=self.light_profile.__class__.__name__,
                ),
            )

    def luminosity_within_circle_in_electrons_per_second_as_function_of_radius(
        self,
        light_profile,
        minimum_radius=1.0e-4,
        maximum_radius=10.0,
        radii_bins=10,
        plot_axis_type="semilogy",
    ):

        radii = plotter_util.quantity_radii_from_minimum_and_maximum_radii_and_radii_points(
            minimum_radius=minimum_radius,
            maximum_radius=maximum_radius,
            radii_points=radii_bins,
        )

        luminosities = list(
            map(
                lambda radius: light_profile.luminosity_within_circle(radius=radius),
                radii,
            )
        )

        self.line(quantity=luminosities, radii=radii, plot_axis_type=plot_axis_type)



from os import path

import autogalaxy as ag
import autogalaxy.plot as aplt
import pytest

directory = path.dirname(path.realpath(__file__))


@pytest.fixture(name="plot_path")
def make_plane_plotter_setup():
    return path.join(
        "{}".format(path.dirname(path.realpath(__file__))), "files", "plots", "plane"
    )


def test__all_individual_plotter__output_file_with_default_name(
    plane_7x7,
    sub_grid_7x7,
    mask_7x7,
    grid_irregular_grouped_7x7,
    include_2d_all,
    plot_path,
    plot_patch,
):

    plane_plotter = aplt.PlanePlotter(
        plane=plane_7x7,
        grid=sub_grid_7x7,
        include_2d=include_2d_all,
        mat_plot_2d=aplt.MatPlot2D(output=aplt.Output(plot_path, format="png")),
    )

    plane_plotter.figures(image=True, plane_image=True, plane_grid=True)

    assert path.join(plot_path, "image.png") in plot_patch.paths
    assert path.join(plot_path, "plane_image.png") in plot_patch.paths
    assert path.join(plot_path, "plane_grid.png") in plot_patch.paths

    plane_7x7.galaxies[0].hyper_galaxy = ag.HyperGalaxy()
    plane_7x7.galaxies[0].hyper_model_image = ag.Array2D.ones(
        shape_native=(7, 7), pixel_scales=0.1
    )
    plane_7x7.galaxies[0].hyper_galaxy_image = ag.Array2D.ones(
        shape_native=(7, 7), pixel_scales=0.1
    )

    plane_plotter.figures(contribution_map=True)

    assert path.join(plot_path, "contribution_map.png") in plot_patch.paths



from os import path
import autogalaxy.plot as aplt


class TestLensingMatObj:
    def test__all_load_from_config_correctly(self):

        light_profile_centres_scatter = aplt.LightProfileCentresScatter()

        assert light_profile_centres_scatter.config_dict["s"] == 1

        mass_profile_centres_scatter = aplt.MassProfileCentresScatter()

        assert mass_profile_centres_scatter.config_dict["s"] == 2

        multiple_images_scatter = aplt.MultipleImagesScatter()

        assert multiple_images_scatter.config_dict["s"] == 3

        critical_curves_plot = aplt.CriticalCurvesPlot()

        assert critical_curves_plot.config_dict["width"] == 4

        caustics_plot = aplt.CausticsPlot()

        assert caustics_plot.config_dict["width"] == 5



from autoconf import conf
import autofit as af

from autoarray.inversion import pixelizations as pix, regularization as reg
from autogalaxy.profiles import (
    light_profiles as lp,
    mass_profiles as mp,
    light_and_mass_profiles as lmp,
)
from autogalaxy.hyper import hyper_data as hd
from autogalaxy import exc

from typing import Union, Optional

import copy


class AbstractSetup:
    def _cls_to_prior_model(self, cls):
        """
        Convert an input class to a `af.PriorModel` so that if a user specifies the models input into a `Setup` not as
        a `PriorModel` (or uses the default inputs which are not `PriorModel`'s) they are converted to a `PriorModel`
        for the pipeline.

        If `None` or a `PriorModel` is input it is not converted to a `PriorModel`.

        Parameters
        ----------
        cls : cls
            The class which is converted to a `PriorModel`' if it is not a `PriorModel`.

        Returns
        -------
        af.PriorModel or None
            The converted class.

        """

        if cls is not None:
            if not isinstance(cls, af.PriorModel):
                return af.PriorModel(cls)

        return cls

    def set_light_prior_model_assertions(
        self,
        bulge_prior_model=None,
        disk_prior_model=None,
        envelope_prior_model=None,
        assert_bulge_sersic_above_disk=True,
        assert_chameleon_core_radius_0_above_core_radius_1=True,
    ):
        """
        This sets a number of assertions on the bugle, disk and envelope prior models which can be customized via
        boolean operators. These assertions are:
        
        1) That the bulge `sersic_index` is above the disk `sersic_index`, if both components are Sersic profiles.
        2) That `core_radius_0` of every Chameleon profile is less than `core_radius_1`.
        
        Parameters
        ----------
        bulge_prior_model : af.PriorModel
            The `PriorModel` used to represent the light distribution of a bulge.
        disk_prior_model : af.PriorModel
            The `PriorModel` used to represent the light distribution of a disk.
        envelope_prior_model : af.PriorModel
            The `PriorModel` used to represent the light distribution of a envelope.
        assert_bulge_sersic_above_disk : bool
            If `True`, the `sersic_index` of the bulge is above that of the disk.
        assert_chameleon_core_radius_0_above_core_radius_1 : bool
            If `True`, `core_radius_0` of every Chameleon profile is above its `core_radius_1` value.

        Returns
        -------
        None
        """

        if assert_bulge_sersic_above_disk:

            self.set_bulge_disk_assertions(
                bulge_prior_model=bulge_prior_model, disk_prior_model=disk_prior_model
            )

        if assert_chameleon_core_radius_0_above_core_radius_1:

            self.set_chameleon_assertions(prior_model=bulge_prior_model)
            self.set_chameleon_assertions(prior_model=disk_prior_model)
            self.set_chameleon_assertions(prior_model=envelope_prior_model)

    def set_bulge_disk_assertions(self, bulge_prior_model, disk_prior_model):
        """
        Sets an assertion on the `bulge_prior_model` and `disk_prior_model` such that the `sersic_index` of the
        bulge is higher than that of the `disk`, if both components are modeled using Sersic profiles.

        Parameters
        ----------
        bulge_prior_model : af.PriorModel
            The `PriorModel` used to represent the light distribution of a bulge.
        disk_prior_model : af.PriorModel
            The `PriorModel` used to represent the light distribution of a disk.

        Returns
        -------
        None
        """

        if bulge_prior_model is not None and disk_prior_model is not None:
            if hasattr(bulge_prior_model, "sersic_index") and hasattr(
                disk_prior_model, "sersic_index"
            ):
                bulge_prior_model.add_assertion(
                    bulge_prior_model.sersic_index > disk_prior_model.sersic_index
                )

    def set_chameleon_assertions(self, prior_model):
        """
        Sets the assertion on all `PriorModels` which are a `Chameleon` profile such that the core radius of the first
        isothermal profile is lower than the second, preventing negative mass.

        Parameters
        ----------
        prior_model : af.PriorModel
            The `PriorModel` that may contain a `Chameleon` profile.

        Returns
        -------
        None
        """
        if prior_model is not None:
            if hasattr(prior_model, "core_radius_0") and hasattr(
                prior_model, "core_radius_1"
            ):
                prior_model.add_assertion(
                    prior_model.core_radius_0 < prior_model.core_radius_1
                )

    @property
    def tag(self):
        raise NotImplementedError


class SetupHyper(AbstractSetup):
    def __init__(
        self,
        hyper_galaxies: bool = False,
        hyper_image_sky: Optional[type(hd.HyperImageSky)] = None,
        hyper_background_noise: Optional[type(hd.HyperBackgroundNoise)] = None,
        hyper_search_no_inversion: Optional[af.NonLinearSearch] = None,
        hyper_search_with_inversion: Optional[af.NonLinearSearch] = None,
        evidence_tolerance: Optional[float] = None,
    ):
        """
        The hyper setup of a pipeline, which controls how hyper-features in PyAutoGalaxy template pipelines run,
        for example controlling whether hyper galaxies are used to scale the noise and the non-linear searches used
        in these phases.

        Users can write their own pipelines which do not use or require the *SetupHyper* class.

        This class enables pipeline tagging, whereby the hyper setup of the pipeline is used in the template pipeline
        scripts to tag the output path of the results depending on the setup parameters. This allows one to fit
        different models to a dataset in a structured path format.

        Parameters
        ----------
        hyper_galaxies : bool
            If a hyper-pipeline is being used, this determines if hyper-galaxy functionality is used to scale the
            noise-map of the dataset throughout the fitting.
        hyper_image_sky : bool
            If a hyper-pipeline is being used, this determines if hyper-galaxy functionality is used include the
            image's background sky component in the model.
        hyper_background_noise : bool
            If a hyper-pipeline is being used, this determines if hyper-galaxy functionality is used include the
            noise-map's background component in the model.
        hyper_search_no_inversion : af.NonLinearSearch or None
            The `NonLinearSearch` used by every inversion phase.
        hyper_search_with_inversion : af.NonLinearSearch or None
            The `NonLinearSearch` used by every hyper combined phase.
        evidence_tolerance : float
            The evidence tolerance of the non-linear searches used in the hyper phases, whereby higher values will
            lead them to end earlier at the expense of accuracy.
        """

        self.evidence_tolerance = evidence_tolerance

        if evidence_tolerance is not None:
            if (
                hyper_search_no_inversion is not None
                or hyper_search_with_inversion is not None
            ):
                raise exc.PipelineException(
                    "You have manually specified a search in the SetupPipeline, and an evidence_tolerance."
                    "You cannot manually specify both - remove one."
                    "(If you want the hyper search to use a specific evidence tolerance, include the evidence"
                    "tolerance in its parameters"
                )

        self.hyper_galaxies = hyper_galaxies

        self.hyper_galaxy_names = None

        if hyper_search_no_inversion is None:
            self.hyper_search_no_inversion = af.DynestyStatic(
                n_live_points=50,
                evidence_tolerance=self.evidence_tolerance,
                sample="rstagger",
            )
        elif hyper_search_no_inversion is not None:
            self.hyper_search_no_inversion = hyper_search_no_inversion

        if hyper_search_with_inversion is None:
            self.hyper_search_with_inversion = af.DynestyStatic(
                n_live_points=50,
                evidence_tolerance=self.evidence_tolerance,
                sample="rstagger",
            )
        else:
            self.hyper_search_with_inversion = hyper_search_with_inversion

        self.hyper_image_sky = hyper_image_sky
        self.hyper_background_noise = hyper_background_noise

    @property
    def component_name(self) -> str:
        """
        The name of the hyper component of a `hyper` pipeline which preceeds the `Setup` tag contained within square
        brackets.

        For the default configuration files in `config/notation/setup_tags.ini` this tag appears as `hyper[tag]`.

        Returns
        -------
        str
            The component name of the hyper pipeline.
        """
        return conf.instance["notation"]["setup_tags"]["names"]["hyper"]

    @property
    def hypers_all_off(self):
        if not self.hyper_galaxies:
            if self.hyper_image_sky is None and self.hyper_background_noise is None:
                return True
        return False

    @property
    def hypers_all_except_image_sky_off(self):
        if not self.hyper_galaxies:
            if self.hyper_background_noise is None:
                return True
        return False

    @property
    def tag(self):
        """
        Tag the pipeline according to the setup of the hyper features, which customizes the pipeline output paths.

        This includes tags for whether hyper-galaxies are used to scale the noise-map and whether the background sky or
        noise are fitted for by the pipeline.

        For the default configuration files in `config/notation/setup_tags.ini` example tags appear as:

        - hyper[galaxies__bg_sky]
        - hyper[bg_sky__bg_noise]
        """
        if self.hypers_all_off:
            return ""

        return (
            f"{self.component_name}["
            f"{self.hyper_galaxies_tag}"
            f"{self.hyper_image_sky_tag}"
            f"{self.hyper_background_noise_tag}]"
        )

    @property
    def hyper_galaxies_tag(self) -> str:
        """
        Tag for if hyper-galaxies are used in a hyper pipeline to scale the noise-map during model fitting, which
        customizes the pipeline's output paths.

        This tag depends on the `hyper_galaxies` bool of the `SetupHyper`.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        - `hyper_galaxies=`False` -> No Tag
        - `hyper_galaxies=`True` -> hyper[galaxies]

        This is used to generate an overall tag in `tag`.
        """
        if not self.hyper_galaxies:
            return ""
        return conf.instance["notation"]["setup_tags"]["hyper"]["hyper_galaxies"]

    @property
    def hyper_image_sky_tag(self) -> str:
        """
        Tag for if the sky-background is a fitted for as a parameter in a hyper pipeline, which customizes the
        pipeline' output paths.

        This tag depends on the `hyper_image_sky` bool of the `SetupHyper`.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        - `hyper_image_sky=`False` -> No Tag
        - `hyper_image_sky=`True` -> hyper[bg_sky]

        This is used to generate an overall tag in `tag`.
        """
        if self.hyper_image_sky is None:
            return ""
        return (
            f"__{conf.instance['notation']['setup_tags']['hyper']['hyper_image_sky']}"
        )

    @property
    def hyper_background_noise_tag(self) -> str:
        """
        Tag for if the background noise level is a fitted for as a parameter in a hyper pipeline, which customizes the
        pipeline' output paths.

        This tag depends on the `hyper_background_noise` bool of the `SetupHyper`.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        - `hyper_background_noise=None` -> No Tag
        - `hyper_background_noise=`True` -> hyper[bg_noise]

        This is used to generate an overall tag in `tag`.
        """
        if self.hyper_background_noise is None:
            return ""
        return f"__{conf.instance['notation']['setup_tags']['hyper']['hyper_background_noise']}"

    def hyper_image_sky_from_result(self, result: af.Result, as_model=True):

        if self.hyper_image_sky is not None:
            if as_model:
                if hasattr(result, "hyper"):
                    return result.hyper.model.hyper_image_sky
                return result.model.hyper_image_sky
            if hasattr(result, "hyper"):
                return result.hyper.instance.hyper_image_sky
            return result.instance.hyper_image_sky

    def hyper_background_noise_from_result(self, result: af.Result):

        if self.hyper_background_noise is not None:
            if hasattr(result, "hyper"):
                return result.hyper.instance.hyper_background_noise
            return result.instance.hyper_background_noise


class AbstractSetupLight(AbstractSetup):
    @property
    def component_name(self) -> str:
        """
        The name of the light component of a `Light` pipeline which preceeds the `Setup` tag contained within square
        brackets.

        For the default configuration files this tag appears as `light[tag]`.

        Returns
        -------
        str
            The component name of the light pipeline.
        """
        return conf.instance["notation"]["setup_tags"]["names"]["light"]


class SetupLightParametric(AbstractSetupLight):
    def __init__(
        self,
        bulge_prior_model: af.PriorModel(lp.LightProfile) = lp.EllipticalSersic,
        disk_prior_model: af.PriorModel(lp.LightProfile) = lp.EllipticalExponential,
        envelope_prior_model: af.PriorModel(lp.LightProfile) = None,
        light_centre: (float, float) = None,
        light_centre_gaussian_prior_values: (float, float) = None,
        align_bulge_disk_centre: bool = True,
        align_bulge_disk_elliptical_comps: bool = False,
        align_bulge_envelope_centre: bool = False,
    ):
        """
        The setup of the parametric light modeling in a pipeline, which controls how PyAutoGalaxy template pipelines
        run, for example controlling assumptions about the bulge-disk model.

        Users can write their own pipelines which do not use or require the `SetupLightParametric` class.

        This class enables pipeline tagging, whereby the setup of the pipeline is used in the template pipeline
        scripts to tag the output path of the results depending on the setup parameters. This allows one to fit
        different models to a dataset in a structured path format.

        Parameters
        ----------
        bulge_prior_model : af.PriorModel(lp.LightProfile)
            The `LightProfile` `PriorModel` used to represent the light distribution of a bulge.
        disk_prior_model : af.PriorModel(lp.LightProfile)
            The `LightProfile` `PriorModel` used to represent the light distribution of a disk.
        envelope_prior_model : af.PriorModel(lp.LightProfile)
            The `LightProfile` `PriorModel` used to represent the light distribution of a envelope.
        light_centre : (float, float) or None
           If input, a fixed (y,x) centre of the galaxy is used for the light profile model which is not treated as a
            free parameter by the non-linear search.
        align_bulge_disk_centre : bool or None
            If a bulge + disk light model (e.g. EllipticalSersic + EllipticalExponential) is used to fit the galaxy,
            `True` will align the centre of the bulge and disk components and not fit them separately.
        align_bulge_disk_elliptical_comps : bool or None
            If a bulge + disk light model (e.g. EllipticalSersic + EllipticalExponential) is used to fit the galaxy,
            `True` will align the elliptical components the bulge and disk components and not fit them separately.
        align_bulge_envelope_centre : bool or None
            If a bulge + envelope light model (e.g. EllipticalSersic + EllipticalExponential) is used to fit the
            galaxy, `True` will align the centre of the bulge and envelope components and not fit them separately.
        """

        self.bulge_prior_model = self._cls_to_prior_model(cls=bulge_prior_model)
        self.disk_prior_model = self._cls_to_prior_model(cls=disk_prior_model)
        self.envelope_prior_model = self._cls_to_prior_model(cls=envelope_prior_model)

        self.light_centre_gaussian_prior_values = light_centre_gaussian_prior_values

        if self.light_centre_gaussian_prior_values is not None:

            mean = light_centre_gaussian_prior_values[0]
            sigma = light_centre_gaussian_prior_values[1]

            if self.bulge_prior_model is not None:

                self.bulge_prior_model.centre_0 = af.GaussianPrior(
                    mean=mean, sigma=sigma
                )
                self.bulge_prior_model.centre_1 = af.GaussianPrior(
                    mean=mean, sigma=sigma
                )

            if self.disk_prior_model is not None:

                self.disk_prior_model.centre_0 = af.GaussianPrior(
                    mean=mean, sigma=sigma
                )
                self.disk_prior_model.centre_1 = af.GaussianPrior(
                    mean=mean, sigma=sigma
                )

            if self.envelope_prior_model is not None:

                self.envelope_prior_model.centre_0 = af.GaussianPrior(
                    mean=mean, sigma=sigma
                )
                self.envelope_prior_model.centre_1 = af.GaussianPrior(
                    mean=mean, sigma=sigma
                )

        self.light_centre = light_centre
        self.align_bulge_disk_centre = align_bulge_disk_centre
        self.align_bulge_disk_elliptical_comps = align_bulge_disk_elliptical_comps

        if self.bulge_prior_model is not None and self.disk_prior_model is not None:

            if self.align_bulge_disk_centre:
                self.bulge_prior_model.centre = self.disk_prior_model.centre

            if self.align_bulge_disk_elliptical_comps:
                if hasattr(self.bulge_prior_model, "elliptical_comps") and hasattr(
                    self.disk_prior_model, "elliptical_comps"
                ):
                    self.bulge_prior_model.elliptical_comps = (
                        self.disk_prior_model.elliptical_comps
                    )

        self.align_bulge_envelope_centre = align_bulge_envelope_centre

        if self.bulge_prior_model is not None and self.envelope_prior_model is not None:

            if self.align_bulge_envelope_centre:
                self.envelope_prior_model.centre = self.bulge_prior_model.centre

        if self.light_centre is not None:

            if self.bulge_prior_model is not None:
                self.bulge_prior_model.centre = self.light_centre

            if self.disk_prior_model is not None:
                self.disk_prior_model.centre = self.light_centre

            if self.envelope_prior_model is not None:
                self.envelope_prior_model.centre = self.light_centre

        self.set_light_prior_model_assertions(
            bulge_prior_model=self.bulge_prior_model,
            disk_prior_model=self.disk_prior_model,
            envelope_prior_model=self.envelope_prior_model,
        )

    @property
    def tag(self):
        """
        Tag the pipeline according to the setup of the light pipeline which customizes the pipeline output paths.

        This includes tags for the `LightProfile` `PriorModel`'s and the alignment of different components in the model.

        For the default configuration files in `config/notation/setup_tags.ini` example tags appear as:

        - light[bulge_sersic__align_bulge_disk_centre]
        - light[bulge_core_sersic__disk_exp__envelope_exp]
        """
        return (
            f"{self.component_name}[parametric"
            f"{self.bulge_prior_model_tag}"
            f"{self.disk_prior_model_tag}"
            f"{self.envelope_prior_model_tag}"
            f"{self.align_bulge_disk_tag}"
            f"{self.align_bulge_envelope_centre_tag}"
            f"{self.light_centre_tag}]"
        )

    @property
    def bulge_prior_model_tag(self) -> str:
        """
        The tag of the bulge `PriorModel` the `LightProfile` class given to the `bulge_prior_model`.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        - `EllipticalSersic` -> sersic
        - `EllipticalExponential` -> exp
        - `SphericalSersic` -> sersic_sph

        Returns
        -------
        str
            The tag of the bulge prior model.
        """

        if self.bulge_prior_model is None:
            return ""

        tag = conf.instance["notation"]["prior_model_tags"]["light"][
            self.bulge_prior_model.name
        ]

        return f"__bulge_{tag}"

    @property
    def disk_prior_model_tag(self) -> str:
        """
        The tag of the disk `PriorModel` the `LightProfile` class given to the `disk_prior_model`.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        - `EllipticalSersic` -> sersic
        - `EllipticalExponential` -> exp
        - `SphericalSersic` -> sersic_sph

        Returns
        -------
        str
            The tag of the disk prior model.
        """

        if self.disk_prior_model is None:
            return ""

        tag = conf.instance["notation"]["prior_model_tags"]["light"][
            self.disk_prior_model.name
        ]

        return f"__disk_{tag}"

    @property
    def envelope_prior_model_tag(self) -> str:
        """
        The tag of the envelope `PriorModel` the `LightProfile` class given to the `envelope_prior_model`.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        - `EllipticalSersic` -> sersic
        - `EllipticalExponential` -> exp
        - `SphericalSersic` -> sersic_sph

        Returns
        -------
        str
            The tag of the envelope prior model.

        """

        if self.envelope_prior_model is None:
            return ""

        tag = conf.instance["notation"]["prior_model_tags"]["light"][
            self.envelope_prior_model.name
        ]

        return f"__envelope_{tag}"

    @property
    def light_centre_tag(self) -> str:
        """
        The tag for whether the centre of the lens light `PriorModel`'s of the pipeline are fixed to an input value
        to customize pipeline output paths.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        light_centre = None -> setup
        light_centre = (1.0, 1.0) -> setup___light_centre_(1.0, 1.0)
        light_centre = (3.0, -2.0) -> setup___light_centre_(3.0, -2.0)
        """
        if self.light_centre is None:
            return ""
        else:
            y = "{0:.2f}".format(self.light_centre[0])
            x = "{0:.2f}".format(self.light_centre[1])
            return f"__{conf.instance['notation']['setup_tags']['light']['light_centre']}_({y},{x})"

    @property
    def align_bulge_disk_centre_tag(self) -> str:
        """
        Tag for if the bulge and disk of a bulge-disk system are aligned or not, to customize pipeline
        output paths based on the bulge-disk model.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        - align_bulge_disk_centre = `False` -> No Tag
        - align_bulge_disk_centre = `True `-> align_bulge_disk_centre
        """
        if self.bulge_prior_model is None or self.disk_prior_model is None:
            return ""

        if not self.align_bulge_disk_centre:
            return ""
        elif self.align_bulge_disk_centre:
            return f"_{conf.instance['notation']['setup_tags']['light']['align_bulge_disk_centre']}"

    @property
    def align_bulge_disk_elliptical_comps_tag(self) -> str:
        """
        Tag if the elliptical components of the bulge and disk `PriorModel`s are aligned or not, to customize pipeline
        output paths based on the bulge-disk model.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        - align_bulge_disk_elliptical_comps = `False` -> No Tag
        - align_bulge_disk_elliptical_comps = `True `-> align_bulge_disk_ell
        """
        if self.bulge_prior_model is None or self.disk_prior_model is None:
            return ""

        if not self.align_bulge_disk_elliptical_comps:
            return ""
        elif self.align_bulge_disk_elliptical_comps:
            return f"_{conf.instance['notation']['setup_tags']['light']['align_bulge_disk_elliptical_comps']}"

    @property
    def align_bulge_envelope_centre_tag(self) -> str:
        """
        Tag for if the bulge and envelope of a bulge-envelope system are aligned or not, to customize pipeline
        output paths based on the bulge-envelope model.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        - align_bulge_envelope_centre = `False` -> No Tag
        - align_bulge_envelope_centre = `True `-> align_bulge_envelope_centre
        """
        if self.bulge_prior_model is None or self.envelope_prior_model is None:
            return ""

        if not self.align_bulge_envelope_centre:
            return ""
        elif self.align_bulge_envelope_centre:
            return f"_{conf.instance['notation']['setup_tags']['light']['align_bulge_envelope_centre']}"

    @property
    def align_bulge_disk_tag(self) -> str:
        """
        Tag the alignment of the the bulge and disk `PriorModel`'s, to customize  pipeline output
        paths based on the bulge-disk model.

        This adds the bulge_disk tags generated in the functions `align_bulge_disk_centre_tag` and
        `align_bulge_disk_elliptical_comps_tag`.

        For the the default configuration files `config/notation/setup_tags.ini` example tags are:

        - align_bulge_disk_ell
        - align_bulge_disk_centre_ell
        """

        align_bulge_disk_tag = f"{self.align_bulge_disk_centre_tag}{self.align_bulge_disk_elliptical_comps_tag}"

        if align_bulge_disk_tag == "":
            return ""

        return f"__{conf.instance['notation']['setup_tags']['light']['align_bulge_disk']}{align_bulge_disk_tag}"


class SetupLightInversion(AbstractSetupLight):
    def __init__(
        self,
        pixelization_prior_model: af.PriorModel(pix.Pixelization),
        regularization_prior_model: af.PriorModel(reg.Regularization),
        inversion_pixels_fixed: float = None,
    ):
        """
        The setup of the inversion light modeling of a pipeline, which controls how PyAutoGalaxy template pipelines run,
        for example controlling the `Pixelization` and `Regularization` used by the `Inversion`.

        Users can write their own pipelines which do not use or require the `SetupLightInversion` class.

        This class enables pipeline tagging, whereby the setup of the pipeline is used in the template pipeline
        scripts to tag the output path of the results depending on the setup parameters. This allows one to fit
        different models to a dataset in a structured path format.

        Parameters
        ----------
        pixelization_prior_model : af.PriorModel(pix.Pixelization)
           If the pipeline uses an `Inversion` to reconstruct the galaxy's light, this determines the `Pixelization`
           used.
        regularization_prior_model : af.PriorModel(reg.Regularization)
            If the pipeline uses an `Inversion` to reconstruct the galaxy's light, this determines the `Regularization`
            scheme used.
        inversion_pixels_fixed : float
            The fixed number of source pixels used by a `Pixelization` class that takes as input a fixed number of
            pixels.
        """

        self._pixelization_prior_model = self._cls_to_prior_model(
            cls=pixelization_prior_model
        )
        self.regularization_prior_model = self._cls_to_prior_model(
            cls=regularization_prior_model
        )

        self.inversion_pixels_fixed = inversion_pixels_fixed

    @property
    def tag(self) -> str:
        """
        Tag the pipeline according to the setup of the `Inversion` used in a light pipeline which customizes the
        pipeline output paths.

        This includes tags for the `Pixelization` `PriorModel` and `Regularization` `PriorModel`.

        For the default configuration files in `config/notation/setup_tags.ini` example tags appear as:

        - light[pix_rect__reg_const]
        - light[pix_voro_image_1200__reg_adapt_bright]
        """
        return f"{self.component_name}[inversion{self.pixelization_tag}{self.inversion_pixels_fixed_tag}{self.regularization_tag}]"

    @property
    def pixelization_prior_model(self) -> af.PriorModel:
        """
        The `PriorModel` of the `Pixelization` in the pipeline.

        This `PriorModel` has its number of pixels fixed to a certain value if the `inversion_pixels_fixed` parameter
        is input.
        """
        if (
            self._pixelization_prior_model.cls is not pix.VoronoiBrightnessImage
            or self.inversion_pixels_fixed is None
        ):
            return self._pixelization_prior_model

        pixelization_prior_model = self._pixelization_prior_model
        pixelization_prior_model.pixels = self.inversion_pixels_fixed
        return pixelization_prior_model

    @property
    def inversion_pixels_fixed_tag(self) -> str:
        """
        The tag for the number of fixed pixels used by the `Inversion` if an fixed input value is used.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        - `inversion_pixels_fixed=None` -> No Tag
        - `inversion_pixels_fixed=1200` -> '_1200'

        Returns
        -------
        str
            The tag of the number of fixed inversion pixels.
        """
        if (
            self.inversion_pixels_fixed is None
            or self._pixelization_prior_model.cls is not pix.VoronoiBrightnessImage
        ):
            return ""

        return f"_{str(self.inversion_pixels_fixed)}"

    @property
    def pixelization_tag(self) -> str:
        """
        The tag for the `Pixelization` `PriorModel` used by the light `Inversion` pipeline.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        `pixelization = af.PriorModel(pix.Rectangular)` -> light[pix_rect]
        `pixelization = af.PriorModel(pix.VoronoiMagnification)` -> light[pix_voro_mag]
        """
        return (
            f"__{conf.instance['notation']['setup_tags']['inversion']['pixelization']}_"
            f"{conf.instance['notation']['prior_model_tags']['pixelization'][self._pixelization_prior_model.name]}"
        )

    @property
    def regularization_tag(self) -> str:
        """
        The tag for the `Regularization` `PriorModel` used by the light `Inversion` pipeline.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        `regularization = reg.Constant` -> light[reg_const]
        `regularization = reg.AdaptiveBrightness` -> light[reg_adapt_bright]
        """
        return (
            f"__{conf.instance['notation']['setup_tags']['inversion']['regularization']}_"
            f"{conf.instance['notation']['prior_model_tags']['regularization'][self.regularization_prior_model.name]}"
        )


class AbstractSetupMass(AbstractSetup):
    def __init__(self, mass_centre: (float, float) = None):
        """
        The setup of the mass modeling in a pipeline for `MassProfile`'s representing the total (e.g. stars + dark
        matter) mass distribution, which controls how PyAutoGalaxy template pipelines run, for example controlling
        assumptions about the bulge-disk model.

        Users can write their own pipelines which do not use or require the `` class.

        This class enables pipeline tagging, whereby the setup of the pipeline is used in the template pipeline
        scripts to tag the output path of the results depending on the setup parameters. This allows one to fit
        different models to a dataset in a structured path format.

        Parameters
        ----------
        mass_centre : (float, float)
           If input, a fixed (y,x) centre of the mass profile is used which is not treated as a free parameter by the
           non-linear search.
        """

        self.mass_centre = mass_centre

    @property
    def component_name(self) -> str:
        """
        The name of the mass component of a `mass` pipeline which preceeds the `Setup` tag contained within square
        brackets.

        For the default configuration files this tag appears as `mass[tag]`.

        Returns
        -------
        str
            The component name of the mass pipeline.
        """
        return conf.instance["notation"]["setup_tags"]["names"]["mass"]

    @property
    def mass_centre_tag(self) -> str:
        """
        The tag for whether the centre of the lens mass `PriorModel`'s of the pipeline are fixed to an input value
        to customize pipeline output paths.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        mass_centre = None -> setup
        mass_centre = (1.0, 1.0) -> setup___mass_centre_(1.0, 1.0)
        mass_centre = (3.0, -2.0) -> setup___mass_centre_(3.0, -2.0)
        """
        if self.mass_centre is None:
            return ""

        y = "{0:.2f}".format(self.mass_centre[0])
        x = "{0:.2f}".format(self.mass_centre[1])
        return f"__{conf.instance['notation']['setup_tags']['mass']['mass_centre']}_({y},{x})"

    def align_centre_to_mass_centre(
        self, mass_prior_model: af.PriorModel(mp.MassProfile)
    ) -> af.PriorModel:
        """
        Align the centre of an input `MassProfile` `PriorModel` to the `mass_centre` of this pipeline setup, such
        that in the model the centre of the mass profile is fixed and not a free parameters that is fitted for.

        If the `mass_centre` is None the mass profile centre is unchanged and remains a model.

        Parameters
        ----------
        mass_prior_model : af.PriorModel(ag.mp.MassProfile)
            The `MassProfile` whose centre may be aligned with the mass_centre attribute.
        """
        if self.mass_centre is not None:
            mass_prior_model.centre = self.mass_centre
        return mass_prior_model

    def unfix_mass_centre(
        self, result: af.Result, mass_prior_model: af.PriorModel(mp.MassProfile)
    ) -> af.PriorModel:
        """
        If the centre of the mass `PriorModel` was previously fixed to an input value via the `mass_centre` input,
        unalign them by making their centre `GaussianPrior`'s with `mean` centred on the input `mass_centre`.

        If `mass_centre` was not input an the centre was fixed in the pipeline itsef, this function can be used to
        unfix the centre to the model result of a previous phase in the pipeline.

        Parameters
        ----------
        result : af.Result
            The result of the previous source or light pipeline.        
        mass_prior_model : af.PriorModel(ag.mp.MassProfile)
            The mass profile whose centre may be unfixed from a previous model.
        """

        if self.mass_centre is not None:

            mass_prior_model.centre.centre_0 = af.GaussianPrior(
                mean=self.mass_centre[0], sigma=0.05
            )
            mass_prior_model.centre.centre_1 = af.GaussianPrior(
                mean=self.mass_centre[1], sigma=0.05
            )

        else:

            mass_prior_model.centre = result.last.model.galaxies.lens.mass.centre

        return mass_prior_model


class SetupMassTotal(AbstractSetupMass):
    def __init__(
        self,
        mass_prior_model: af.PriorModel(mp.MassProfile) = mp.EllipticalPowerLaw,
        mass_centre: (float, float) = None,
        align_bulge_mass_centre: bool = False,
    ):
        """
        The setup of the mass modeling in a pipeline for `MassProfile`'s representing the total (e.g. stars + dark
        matter) mass distribution, which controls how PyAutoGalaxy template pipelines run, for example controlling
        assumptions about the bulge-disk model.

        Users can write their own pipelines which do not use or require the `SetupMassTotal` class.

        This class enables pipeline tagging, whereby the setup of the pipeline is used in the template pipeline
        scripts to tag the output path of the results depending on the setup parameters. This allows one to fit
        different models to a dataset in a structured path format.

        Parameters
        ----------
        mass_prior_model : af.PriorModel(mp.MassProfile)
            The `MassProfile` fitted by the `Pipeline` (the pipeline must specifically use this option to use this
            mass profile)
        mass_centre : (float, float) or None
           If input, a fixed (y,x) centre of the mass profile is used which is not treated as a free parameter by the
           non-linear search.
        align_bulge_mass_centre : bool
            If `True` and the galaxy model has both a light and mass component, the function
            `align_centre_of_mass_to_light` can be used to align their centres.
        """

        super().__init__(mass_centre=mass_centre)

        self.mass_prior_model = self._cls_to_prior_model(cls=mass_prior_model)

        self.align_bulge_mass_centre = align_bulge_mass_centre

        if self.mass_centre is not None:
            self.mass_prior_model.centre = self.mass_centre

    @property
    def tag(self):
        """
        Tag the pipeline according to the setup of the total mass pipeline which customizes the pipeline output paths.

        This includes tags for the `MassProfile` `PriorModel`'s and the alignment of different components in the model.

        For the default configuration files in `config/notation/setup_tags.ini` example tags appear as:

        - mass[total__sie]
        - mass[total__power_law__centre_(0.0,0.0)]
        """
        return (
            f"{self.component_name}[total"
            f"{self.mass_prior_model_tag}"
            f"{self.mass_centre_tag}"
            f"{self.align_bulge_mass_centre_tag}]"
        )

    @property
    def mass_prior_model_tag(self) -> str:
        """
        The tag of the mass `PriorModel` the `MassProfile` class given to the `mass_prior_model`.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        - `EllipticalIsothermal` -> sie
        - `EllipticalPowerLaw` -> power_law

        Returns
        -------
        str
            The tag of the mass prior model.
        """

        if self.mass_prior_model is None:
            return ""

        return f"__{conf.instance['notation']['prior_model_tags']['mass'][self.mass_prior_model.name]}"

    @property
    def align_bulge_mass_centre_tag(self) -> str:
        """
        Tags if the lens mass model centre is aligned with that of its light profile.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        align_bulge_mass_centre = `False` -> setup
        align_bulge_mass_centre = `True` -> setup___align_bulge_mass_centre
        """
        if self.mass_centre is not None:
            return ""

        if not self.align_bulge_mass_centre:
            return ""
        return f"__{conf.instance['notation']['setup_tags']['mass']['align_bulge_mass_centre']}"

    def align_centre_of_mass_to_light(
        self,
        mass_prior_model: af.PriorModel(mp.MassProfile),
        light_centre: (float, float),
    ):
        """Align the centre of a mass profile to the centre of a light profile, if the align_bulge_mass_centre
        SLaM setting is True.

        Parameters
        ----------
        mass_prior_model : af.PriorModel(ag.mp.MassProfile)
            The mass profile whose centre may be aligned with the light_centre attribute.
        light : (float, float)
            The centre of the light profile the mass profile is aligned with.
        """
        if self.align_bulge_mass_centre:
            mass_prior_model.centre = light_centre
        else:
            mass_prior_model.centre.centre_0 = af.GaussianPrior(
                mean=light_centre[0], sigma=0.1
            )
            mass_prior_model.centre.centre_1 = af.GaussianPrior(
                mean=light_centre[1], sigma=0.1
            )
        return mass_prior_model

    def unalign_mass_centre_from_light_centre(
        self,
        results: af.ResultsCollection,
        mass_prior_model: af.PriorModel(mp.MassProfile),
    ):
        """If the centre of a mass model was previously aligned with that of the lens light centre, unaligned them
        by using an earlier model of the light.

        Parameters
        ----------
        mass_prior_model : af.PriorModel(ag.mp.MassProfile)
            The `MassProfile` whose centre may be aligned with the `LightProfile` centre.
        """
        if self.align_bulge_mass_centre:

            mass_prior_model.centre = results[-3].model.galaxies.lens.bulge.centre

        else:

            mass_prior_model.centre = results.last.model.galaxies.lens.mass.centre

        return mass_prior_model

    def mass_prior_model_with_updated_priors_from_result(
        self, result: af.Result, unfix_mass_centre=False
    ):
        """
        Returns an updated version of the `mass_prior_model` whose priors are initialized from previous results in a
        pipeline.

        This function generically links any `MassProfile` to any `MassProfile`, pairing parameters which share the
        same path.

        Parameters
        ----------
        results : af.ResultsCollection
            The results of the previous mass pipeline.

        Returns
        -------
        af.PriorModel(mp.MassProfile)
            The total mass profile whose priors are initialized from a previous result.
        """

        mass = self._cls_to_prior_model(cls=self.mass_prior_model.cls)

        mass.take_attributes(source=result.model.galaxies.lens.mass)

        if unfix_mass_centre and isinstance(mass.centre, tuple):

            centre_tuple = mass.centre

            mass.centre = self._cls_to_prior_model(cls=self.mass_prior_model.cls).centre

            mass.centre.centre_0 = af.GaussianPrior(mean=centre_tuple[0], sigma=0.05)
            mass.centre.centre_1 = af.GaussianPrior(mean=centre_tuple[1], sigma=0.05)

        return mass


class SetupMassLightDark(AbstractSetupMass):
    def __init__(
        self,
        bulge_prior_model: af.PriorModel(lmp.LightMassProfile) = lmp.EllipticalSersic,
        disk_prior_model: af.PriorModel(
            lmp.LightMassProfile
        ) = lmp.EllipticalExponential,
        envelope_prior_model: af.PriorModel(lmp.LightMassProfile) = None,
        dark_prior_model: af.PriorModel(mp.MassProfile) = mp.EllipticalNFWMCRLudlow,
        mass_centre: (float, float) = None,
        constant_mass_to_light_ratio: bool = False,
        align_bulge_dark_centre: bool = False,
    ):
        """
        The setup of the mass modeling in a pipeline for `MassProfile`'s representing the decomposed light and dark
        mass distributions, which controls how PyAutoGalaxy template pipelines run, for example controlling assumptions
        about the bulge-disk model.

        Users can write their own pipelines which do not use or require the `SetupMassLightDark` class.

        This class enables pipeline tagging, whereby the setup of the pipeline is used in the template pipeline
        scripts to tag the output path of the results depending on the setup parameters. This allows one to fit
        different models to a dataset in a structured path format.

        Parameters
        ----------
        bulge_prior_model : af.PriorModel(lmp.LightMassProfile)
            The `LightMassProfile` `PriorModel` used to represent the light and mass distribution of the bulge.
        disk_prior_model : af.PriorModel(lmp.LightMassProfile)
            The `LightMassProfile` `PriorModel` used to represent the light and mass distribution of the disk.
        envelope_prior_model : af.PriorModel(lmp.LightMassProfile)
            The `LightMassProfile` `PriorModel` used to represent the light and mass distribution of the stellar envelope.
        dark_prior_model : af.PriorModel(mp.MassProfile)
            The `MassProfile` `PriorModel` used to represent the dark matter distribution of the dark matter halo.
        mass_centre : (float, float)
           If input, a fixed (y,x) centre of the mass profile is used which is not treated as a free parameter by the
           non-linear search.
        constant_mass_to_light_ratio : bool
            If True, and the mass model consists of multiple `LightProfile` and `MassProfile` coomponents, the
            mass-to-light ratio's of all components are fixed to one shared value.
        align_bulge_mass_centre : bool
            If True, and the mass model is a decomposed bulge, disk and dark matter model (e.g. EllipticalSersic +
            EllipticalExponential + SphericalNFW), the centre of the bulge and dark matter profiles are aligned.
        """

        super().__init__(mass_centre=mass_centre)

        self.bulge_prior_model = self._cls_to_prior_model(cls=bulge_prior_model)
        self.disk_prior_model = self._cls_to_prior_model(cls=disk_prior_model)
        self.envelope_prior_model = self._cls_to_prior_model(cls=envelope_prior_model)

        self.dark_prior_model = self._cls_to_prior_model(cls=dark_prior_model)

        self.constant_mass_to_light_ratio = constant_mass_to_light_ratio
        self.align_bulge_dark_centre = align_bulge_dark_centre

        if self.align_bulge_dark_centre:
            self.dark_prior_model.centre = self.bulge_prior_model.centre

        if self.constant_mass_to_light_ratio:
            for profile in self.light_and_mass_prior_models[1:]:
                profile.mass_to_light_ratio = self.light_and_mass_prior_models[
                    0
                ].mass_to_light_ratio

    @property
    def light_and_mass_prior_models(self):
        return list(
            filter(
                None,
                [
                    self.bulge_prior_model,
                    self.disk_prior_model,
                    self.envelope_prior_model,
                ],
            )
        )

    @property
    def tag(self):
        """
        Tag the pipeline according to the setup of the decomposed light and dark mass pipeline which customizes
        the pipeline output paths.

        This includes tags for the `MassProfile` `PriorModel`'s and the alignment of different components in the model.

        For the default configuration files in `config/notation/setup_tags.ini` example tags appear as:

        - mass[light_dark__bulge_]
        - mass[total_power_law__centre_(0.0,0.0)]
        """
        return (
            f"{self.component_name}[light_dark"
            f"{self.bulge_prior_model_tag}"
            f"{self.disk_prior_model_tag}"
            f"{self.envelope_prior_model_tag}"
            f"{self.constant_mass_to_light_ratio_tag}"
            f"{self.dark_prior_model_tag}"
            f"{self.mass_centre_tag}"
            f"{self.align_bulge_dark_centre_tag}]"
        )

    @property
    def constant_mass_to_light_ratio_tag(self) -> str:
        """Generate a tag for whether the mass-to-light ratio in a light-dark mass model is constaant (shared amongst
         all light and mass profiles) or free (all mass-to-light ratios are free parameters).

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        constant_mass_to_light_ratio = `False` -> mlr_free
        constant_mass_to_light_ratio = `True` -> mlr_constant
        """
        if self.constant_mass_to_light_ratio:
            return f"__mlr_{conf.instance['notation']['setup_tags']['mass']['constant_mass_to_light_ratio']}"
        return f"__mlr_{conf.instance['notation']['setup_tags']['mass']['free_mass_to_light_ratio']}"

    @property
    def bulge_prior_model_tag(self) -> str:
        """
        The tag of the bulge `PriorModel` using the tags specified in the setup_tags.ini config file.

        This tag depends on the `LightMassProfile` class given to the bulge, for example for the default configuration
        files:

        - `EllipticalSersic` -> sersic
        - `EllipticalExponential` -> exp
        - `SphericalSersic` -> sersic_sph

        Returns
        -------
        str
            The tag of the bulge prior model.

        """

        if self.bulge_prior_model is None:
            return ""

        tag = conf.instance["notation"]["prior_model_tags"]["mass"][
            self.bulge_prior_model.name
        ]

        return f"__bulge_{tag}"

    @property
    def disk_prior_model_tag(self) -> str:
        """
        The tag of the disk `PriorModel` using the tags specified in the setup_tags.ini config file.

        This tag depends on the `LightMassProfile` class given to the disk, for example for the default configuration
        files:

        - `EllipticalSersic` -> sersic
        - `EllipticalExponential` -> exp
        - `SphericalSersic` -> sersic_sph

        Returns
        -------
        str
            The tag of the disk prior model.
        """

        if self.disk_prior_model is None:
            return ""

        tag = conf.instance["notation"]["prior_model_tags"]["mass"][
            self.disk_prior_model.name
        ]

        return f"__disk_{tag}"

    @property
    def envelope_prior_model_tag(self) -> str:
        """
        The tag of the envelope `PriorModel` using the tags specified in the setup_tags.ini config file.

        This tag depends on the `LightMassProfile` class given to the envelope, for example for the default
        configuration files:

        - `EllipticalSersic` -> sersic
        - `EllipticalExponential` -> exp
        - `SphericalSersic` -> sersic_sph

        Returns
        -------
        str
            The tag of the envelope prior model.

        """

        if self.envelope_prior_model is None:
            return ""

        tag = conf.instance["notation"]["prior_model_tags"]["mass"][
            self.envelope_prior_model.name
        ]

        return f"__envelope_{tag}"

    @property
    def dark_prior_model_tag(self) -> str:
        """
        The tag of the dark `PriorModel` using the tags specified in the setup_tags.ini config file.

        This tag depends on the `MassProfile` class given to the dark, for example for the default
        configuration files:

        - `SphericalNFWMCRLudlow` -> nfw_sph_ludlow
        - `EllipticalNFW` -> nfw
        - `SphericalTruncatedNFW` -> nfw_trun_sph

        Returns
        -------
        str
            The tag of the dark prior model.
        """

        if self.dark_prior_model is None:
            return ""

        tag = conf.instance["notation"]["prior_model_tags"]["mass"][
            self.dark_prior_model.name
        ]

        return f"__dark_{tag}"

    @property
    def align_bulge_dark_centre_tag(self) -> str:
        """
        Tag for if the lens mass model's bulge `PriorModel` is aligned with the dark matter centre.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        align_bulge_dark_centre = `False` -> mass[light_dark]
        align_bulge_dark_centre = `True` -> mass[light_dark__align_bulge_dark_centre]
        """
        if not self.align_bulge_dark_centre:
            return ""
        return f"__{conf.instance['notation']['setup_tags']['mass']['align_bulge_dark_centre']}"

    def align_bulge_and_dark_centre(self, bulge_prior_model, dark_prior_model):
        """
        Align the centre of input bulge `PriorModel` with that of the `PriorModel` representing the dark `MassProfile`,
        depending on the `align_bulge_darl_centre` attribute of the `SetupMassLightDark` instance.

        Parameters
        ----------
        bulge_prior_model : af.PriorModel(ag.lmp.LightMassProfile)
            The `LightMassProfile` representing the bulge whose centre is aligned with that of the dark matter.
        disk_prior_model : af.PriorModel(ag.lmp.LightMassProfile)
            The `LightMassProfile` representing the dark matter whose centre is aligned with that of the bulge.
        """
        if self.align_bulge_dark_centre:
            dark_prior_model.centre = bulge_prior_model.centre

    def light_and_mass_prior_models_with_updated_priors(
        self, result: af.Result, einstein_mass_range=None, as_instance=False
    ):
        """
        Returns an updated version of the `bulge_prior_model`, `disk_prior_model`_ and `envelope_prior_model`,  whose
        priors are initialized from previous results of a `Light` pipeline.

        This function generically links any `LightProfile` to any `LightProfile`, pairing parameters which share the
        same path

        Parameters
        ----------
        result : af.ResultsCollection
            The results of the previous source pipeline.

        Returns
        -------
        af.PriorModel(mp.MassProfile)
            The light profiles whose priors are initialized from a previous result.
        """

        if self.bulge_prior_model is None:
            bulge = None
        else:
            bulge = copy.deepcopy(self.bulge_prior_model)

            if as_instance:
                bulge.take_attributes(source=result.instance.galaxies.lens.bulge)
            else:
                bulge.take_attributes(source=result.model.galaxies.lens.bulge)

            if einstein_mass_range is not None:

                bulge = self.update_stellar_mass_priors_from_result(
                    prior_model=bulge,
                    result=result,
                    einstein_mass_range=einstein_mass_range,
                )

        if self.disk_prior_model is None:
            disk = None
        else:
            disk = copy.deepcopy(self.disk_prior_model)

            if as_instance:
                disk.take_attributes(source=result.instance.galaxies.lens.disk)
            else:
                disk.take_attributes(source=result.model.galaxies.lens.disk)

            if einstein_mass_range is not None:
                disk = self.update_stellar_mass_priors_from_result(
                    prior_model=disk,
                    result=result,
                    einstein_mass_range=einstein_mass_range,
                )

        if self.envelope_prior_model is None:
            envelope = None
        else:
            envelope = copy.deepcopy(self.envelope_prior_model)

            if as_instance:
                envelope.take_attributes(source=result.instance.galaxies.lens.envelope)
            else:
                envelope.take_attributes(source=result.model.galaxies.lens.envelope)

            if einstein_mass_range is not None:
                envelope = self.update_stellar_mass_priors_from_result(
                    prior_model=envelope,
                    result=result,
                    einstein_mass_range=einstein_mass_range,
                )

        ### TODO : Assertiosn must be after take attributwes, hence this.

        self.set_light_prior_model_assertions(
            bulge_prior_model=bulge,
            disk_prior_model=disk,
            envelope_prior_model=envelope,
        )

        return bulge, disk, envelope

    def update_stellar_mass_priors_from_result(
        self, prior_model, result: af.Result, einstein_mass_range, bins=100
    ):

        if prior_model is None:
            return None

        grid = result.max_log_likelihood_fit.grid

        einstein_radius = result.max_log_likelihood_tracer.einstein_radius_from_grid(
            grid=grid
        )

        einstein_mass = result.max_log_likelihood_tracer.einstein_mass_angular_from_grid(
            grid=grid
        )

        einstein_mass_lower = einstein_mass_range[0] * einstein_mass
        einstein_mass_upper = einstein_mass_range[1] * einstein_mass

        instance = prior_model.instance_from_prior_medians()

        mass_to_light_ratio_lower = instance.normalization_from_mass_angular_and_radius(
            mass_angular=einstein_mass_lower, radius=einstein_radius, bins=bins
        )
        mass_to_light_ratio_upper = instance.normalization_from_mass_angular_and_radius(
            mass_angular=einstein_mass_upper, radius=einstein_radius, bins=bins
        )

        prior_model.mass_to_light_ratio = af.LogUniformPrior(
            lower_limit=mass_to_light_ratio_lower, upper_limit=mass_to_light_ratio_upper
        )

        return prior_model


class SetupSMBH(AbstractSetup):
    def __init__(
        self,
        smbh_prior_model: af.PriorModel(mp.MassProfile) = mp.PointMass,
        smbh_centre_fixed: bool = True,
    ):
        """
        The setup of a super massive black hole (SMBH) in the mass model of a PyAutoGalaxy template pipeline.

        Users can write their own pipelines which do not use or require the *SetupSMBH* class.

        This class enables pipeline tagging, whereby the setup of the pipeline is used in the template pipeline
        scripts to tag the output path of the results depending on the setup parameters. This allows one to fit
        different models to a dataset in a structured path format.

        Parameters
        ----------
        smbh_prior_model : af.PriorModel(mp.MassProfile)
            The `MassProfile` used to model the mass of the super massive black hole.
        smbh_centre_fixed : bool
            If True, the super-massive black hole's centre is fixed to a value input by the pipeline, else it is
            free to vary in the model.
        """
        self.smbh_prior_model = self._cls_to_prior_model(cls=smbh_prior_model)
        self.smbh_centre_fixed = smbh_centre_fixed

    @property
    def component_name(self) -> str:
        """
        The name of the smbh component of a `smbh` pipeline which preceeds the `Setup` tag contained within square
        brackets.

        For the default configuration files this tag appears as `smbh[tag]`.

        Returns
        -------
        str
            The component name of the smbh pipeline.
        """
        return conf.instance["notation"]["setup_tags"]["names"]["smbh"]

    @property
    def tag(self):
        return (
            f"{self.component_name}[{self.smbh_prior_model_tag}{self.smbh_centre_tag}]"
        )

    @property
    def smbh_prior_model_tag(self) -> str:
        """
        The tag of the smbh `PriorModel` the `MassProfile` class given to the `smbh_prior_model`.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        - `EllipticalIsothermal` -> sie
        - `EllipticalPowerLaw` -> power_law

        Returns
        -------
        str
            The tag of the smbh prior model.
        """
        return f"{conf.instance['notation']['prior_model_tags']['mass'][self.smbh_prior_model.name]}"

    @property
    def smbh_centre_tag(self) -> str:
        """
        Tag whether the smbh `PriorModel` centre is fixed or fitted for as a free parameter.

        For the the default configuration files `config/notation/setup_tags.ini` tagging is performed as follows:

        smbh_centre_fixed=True -> smbh[centre_fixed]
        smbh_centre_fixed=False -> smbh[centre_free]
        """

        if self.smbh_centre_fixed:

            smbh_centre_tag = conf.instance["notation"]["setup_tags"]["smbh"][
                "smbh_centre_fixed"
            ]

        else:

            smbh_centre_tag = conf.instance["notation"]["setup_tags"]["smbh"][
                "smbh_centre_free"
            ]

        return f"__{smbh_centre_tag}"

    def smbh_from_centre(self, centre, centre_sigma=0.1) -> af.PriorModel:
        """
        Returns a `PriorModel` of the `smbh_prior_model` which is fitted for in the mass-model representing a
        super-massive black-hole (smbh).

        The centre of the smbh is an input parameter of the function, and this centre is either fixed to the input
        values as an instance or fitted for as a model.

        Parameters
        ----------
        centre : (float, float)
            The centre of the `PointMass` that repreents the super-massive black hole.
        centre_sigma : float
            If the centre is free, this is the sigma value of each centre's _GaussianPrior_.
        """

        smbh = af.PriorModel(mp.PointMass)

        if self.smbh_centre_fixed:
            smbh.centre = centre
        else:
            smbh.centre.centre_0 = af.GaussianPrior(mean=centre[0], sigma=centre_sigma)
            smbh.centre.centre_1 = af.GaussianPrior(mean=centre[1], sigma=centre_sigma)

        return smbh


class SetupPipeline:
    def __init__(
        self,
        path_prefix: str = None,
        redshift_galaxy: float = 1.0,
        setup_hyper: SetupHyper = None,
        setup_light: Union[SetupLightParametric, SetupLightInversion] = None,
        setup_mass: Union[SetupMassTotal, SetupMassLightDark] = None,
        setup_smbh: SetupSMBH = None,
    ):
        """
        The setup of a pipeline, which controls how PyAutoGalaxy template pipelines runs, for example controlling
        assumptions about the light and mass models.

        Users can write their own pipelines which do not use or require the *SetupPipeline* class.

        This class enables pipeline tagging, whereby the setup of the pipeline is used in the template pipeline
        scripts to tag the output path of the results depending on the setup parameters. This allows one to fit
        different models to a dataset in a structured path format.

        Parameters
        ----------
        path_prefix : str or None
            The prefix of folders between the output path of the pipeline and the pipeline name, tags and phase folders.
        redshift_galaxy : float
            The redshift of the galaxy fitted, used by the pipeline for converting arc-seconds to kpc, masses to
            solMass, etc.
        setup_hyper : SetupHyper
            The setup of the hyper analysis if used (e.g. hyper-galaxy noise scaling).
        setup_light : SetupLightParametric or SetupLightInversion
            The setup of the light profile modeling (e.g. for bulge-disk models if they are geometrically aligned).
        setup_mass : SetupMassTotal or SetupMassLighDark
            The setup of the mass modeling (e.g. if a constant mass to light ratio is used).
        setup_smbh : SetupSMBH
            The setup of the super-massive black hole modeling (e.g. its `MassProfile` and if its centre is fixed).
        """

        self.path_prefix = path_prefix
        self.redshift_galaxy = redshift_galaxy
        self.setup_hyper = setup_hyper
        self.setup_light = setup_light
        self.setup_mass = setup_mass
        self.setup_smbh = setup_smbh

        if isinstance(self.setup_light, SetupLightParametric) and isinstance(
            self.setup_mass, SetupMassLightDark
        ):
            self.setup_mass.bulge_prior_model = self.setup_light.bulge_prior_model
            self.setup_mass.disk_prior_model = self.setup_light.disk_prior_model
            self.setup_mass_envelope_prior_model = self.setup_light.envelope_prior_model

    def _pipeline_tag_from_setup(self, setup: AbstractSetup) -> str:
        """
        Returns the tag of a pipeline from the tag of an indiviual `Setup` object.

        Parameters
        ----------
        setup_tag : str
            The tag of the `Setup` object which is converted to the pipeline tag.

        Returns
        -------
        str
            The pipeline tag.
        """
        return f"__{setup.tag}" if setup is not None else ""

    @property
    def tag(self) -> str:
        """
        The overall pipeline tag, which customizes the 'setup' folder the results are output to.

        For the the default configuration files `config/notation/setup_tags.ini` examples of tagging are as follows:

        - setup__hyper[galaxies__bg_noise]__light[bulge_sersic__disk__exp_light_centre_(1.00,2.00)]
        - "setup__smbh[point_mass__centre_fixed]"
        """

        setup_tag = conf.instance["notation"]["setup_tags"]["pipeline"]["pipeline"]

        hyper_tag = self._pipeline_tag_from_setup(setup=self.setup_hyper)
        light_tag = self._pipeline_tag_from_setup(setup=self.setup_light)
        mass_tag = self._pipeline_tag_from_setup(setup=self.setup_mass)
        smbh_tag = self._pipeline_tag_from_setup(setup=self.setup_smbh)

        return f"{setup_tag}{hyper_tag}{light_tag}{mass_tag}{smbh_tag}"



from os import path

import autogalaxy.plot as aplt
import pytest


@pytest.fixture(name="plot_path")
def make_fit_interferometer_plotter_setup():
    return path.join(
        "{}".format(path.dirname(path.realpath(__file__))), "files", "plots", "fit"
    )


def test__fit_sub_plot_real_space(
    masked_interferometer_fit_7x7,
    masked_interferometer_fit_x2_galaxy_inversion_7x7,
    include_2d_all,
    plot_path,
    plot_patch,
):

    fit_interferometer_plotter = aplt.FitInterferometerPlotter(
        fit=masked_interferometer_fit_7x7,
        include_2d=include_2d_all,
        mat_plot_2d=aplt.MatPlot2D(output=aplt.Output(plot_path, format="png")),
    )

    fit_interferometer_plotter.subplot_fit_real_space()

    assert path.join(plot_path, "subplot_fit_real_space.png") in plot_patch.paths

    plot_patch.paths = []

    fit_interferometer_plotter = aplt.FitInterferometerPlotter(
        fit=masked_interferometer_fit_x2_galaxy_inversion_7x7,
        include_2d=include_2d_all,
        mat_plot_2d=aplt.MatPlot2D(output=aplt.Output(plot_path, format="png")),
    )

    fit_interferometer_plotter.subplot_fit_real_space()

    assert path.join(plot_path, "subplot_fit_real_space.png") in plot_patch.paths



from autoarray.mask import mask_2d
from autoarray.structures import arrays, grids, vector_fields
from autoarray.plot.mat_wrap import visuals as vis

from matplotlib import patches as ptch
import typing
from typing import List


class Visuals1D(vis.Visuals1D):

    pass


class Visuals2D(vis.Visuals2D):
    def __init__(
        self,
        origin: grids.Grid2D = None,
        border: grids.Grid2D = None,
        mask: mask_2d.Mask2D = None,
        positions: grids.Grid2DIrregular = None,
        grid: grids.Grid2D = None,
        pixelization_grid: grids.Grid2D = None,
        vector_field: vector_fields.VectorField2DIrregular = None,
        patches: typing.Union[ptch.Patch] = None,
        array_overlay: arrays.Array2D = None,
        light_profile_centres: grids.Grid2DIrregular = None,
        mass_profile_centres: grids.Grid2DIrregular = None,
        multiple_images: grids.Grid2DIrregular = None,
        critical_curves: grids.Grid2DIrregular = None,
        caustics: grids.Grid2DIrregular = None,
        indexes: typing.Union[List[int], List[List[int]]] = None,
        pixelization_indexes: typing.Union[List[int], List[List[int]]] = None,
    ):

        super().__init__(
            mask=mask,
            positions=positions,
            grid=grid,
            pixelization_grid=pixelization_grid,
            vector_field=vector_field,
            patches=patches,
            array_overlay=array_overlay,
            origin=origin,
            border=border,
            indexes=indexes,
            pixelization_indexes=pixelization_indexes,
        )

        self.light_profile_centres = light_profile_centres
        self.mass_profile_centres = mass_profile_centres
        self.multiple_images = multiple_images
        self.critical_curves = critical_curves
        self.caustics = caustics

    def plot_via_plotter(self, plotter, grid_indexes=None, mapper=None):

        super().plot_via_plotter(
            plotter=plotter, grid_indexes=grid_indexes, mapper=mapper
        )

        if self.light_profile_centres is not None:
            plotter.light_profile_centres_scatter.scatter_grid_grouped(
                grid_grouped=self.light_profile_centres
            )

        if self.mass_profile_centres is not None:
            plotter.mass_profile_centres_scatter.scatter_grid_grouped(
                grid_grouped=self.mass_profile_centres
            )

        if self.multiple_images is not None:
            plotter.multiple_images_scatter.scatter_grid_grouped(
                grid_grouped=self.multiple_images
            )

        if self.critical_curves is not None:
            plotter.critical_curves_plot.plot_grid_grouped(
                grid_grouped=self.critical_curves
            )

        if self.caustics is not None:
            plotter.caustics_plot.plot_grid_grouped(grid_grouped=self.caustics)



from astropy.io import fits

import os
from os import path
import numpy as np
import autogalaxy as ag


def create_fits(fits_path, array):

    if path.exists(fits_path):
        os.remove(fits_path)

    hdu_list = fits.HDUList()

    hdu_list.append(fits.ImageHDU(array))

    hdu_list.writeto(f"{fits_path}")


class TestImaging:
    def test__from_fits__all_imaging_data_structures_are_flipped_for_ds9(self):

        fits_path = path.join(
            "{}".format(path.dirname(path.realpath(__file__))), "files"
        )
        image_path = path.join(fits_path, "image.fits")

        create_fits(fits_path=image_path, array=[[1.0, 0.0], [0.0, 0.0]])

        noise_map_path = path.join(fits_path, "noise_map.fits")
        create_fits(fits_path=noise_map_path, array=[[2.0, 0.0], [0.0, 0.0]])

        psf_path = path.join(fits_path, "psf.fits")
        create_fits(fits_path=psf_path, array=[[1.0, 1.0], [0.0, 0.0]])

        imaging = ag.Imaging.from_fits(
            image_path=image_path,
            noise_map_path=noise_map_path,
            psf_path=psf_path,
            pixel_scales=0.1,
        )

        assert (imaging.image.native == np.array([[0.0, 0.0], [1.0, 0.0]])).all()
        assert (imaging.noise_map.native == np.array([[0.0, 0.0], [2.0, 0.0]])).all()
        assert (imaging.psf.native == np.array([[0.0, 0.0], [0.5, 0.5]])).all()

        imaging.output_to_fits(
            image_path=image_path,
            noise_map_path=noise_map_path,
            psf_path=psf_path,
            overwrite=True,
        )

        hdu_list = fits.open(image_path)
        image = np.array(hdu_list[0].data).astype("float64")
        assert (image == np.array([[1.0, 0.0], [0.0, 0.0]])).all()

        hdu_list = fits.open(noise_map_path)
        noise_map = np.array(hdu_list[0].data).astype("float64")
        assert (noise_map == np.array([[2.0, 0.0], [0.0, 0.0]])).all()

        hdu_list = fits.open(psf_path)
        psf = np.array(hdu_list[0].data).astype("float64")
        assert (psf == np.array([[0.5, 0.5], [0.0, 0.0]])).all()


class TestMaskedImaging:
    def test__masked_dataset_via_autoarray(self, imaging_7x7, sub_mask_7x7):

        masked_imaging_7x7 = ag.MaskedImaging(imaging=imaging_7x7, mask=sub_mask_7x7)

        assert (masked_imaging_7x7.image.slim == np.ones(9)).all()

        assert (
            masked_imaging_7x7.image.native == np.ones((7, 7)) * np.invert(sub_mask_7x7)
        ).all()

        assert (masked_imaging_7x7.noise_map.slim == 2.0 * np.ones(9)).all()
        assert (
            masked_imaging_7x7.noise_map.native
            == 2.0 * np.ones((7, 7)) * np.invert(sub_mask_7x7)
        ).all()

        assert (masked_imaging_7x7.psf.slim == (1.0 / 9.0) * np.ones(9)).all()
        assert (masked_imaging_7x7.psf.native == (1.0 / 9.0) * np.ones((3, 3))).all()

        assert type(masked_imaging_7x7.convolver) == ag.Convolver

    def test__inheritance_from_autoarray(
        self, imaging_7x7, sub_mask_7x7, blurring_grid_7x7
    ):

        masked_imaging_7x7 = ag.MaskedImaging(
            imaging=imaging_7x7,
            mask=sub_mask_7x7,
            settings=ag.SettingsMaskedImaging(
                grid_class=ag.Grid2D, psf_shape_2d=(3, 3)
            ),
        )

        grid = ag.Grid2D.from_mask(mask=sub_mask_7x7)

        assert (masked_imaging_7x7.grid == grid).all()

        blurring_grid = grid.blurring_grid_from_kernel_shape(kernel_shape_native=(3, 3))

        assert (masked_imaging_7x7.blurring_grid.slim == blurring_grid_7x7).all()
        assert (masked_imaging_7x7.blurring_grid == blurring_grid).all()

    def test__modified_image_and_noise_map(
        self, image_7x7, noise_map_7x7, imaging_7x7, sub_mask_7x7
    ):

        masked_imaging_7x7 = ag.MaskedImaging(imaging=imaging_7x7, mask=sub_mask_7x7)

        image_7x7[0] = 10.0
        noise_map_7x7[0] = 11.0

        masked_imaging_7x7 = masked_imaging_7x7.modify_image_and_noise_map(
            image=image_7x7, noise_map=noise_map_7x7
        )

        assert masked_imaging_7x7.image.slim[0] == 10.0
        assert masked_imaging_7x7.image.native[0, 0] == 10.0
        assert masked_imaging_7x7.noise_map.slim[0] == 11.0
        assert masked_imaging_7x7.noise_map.native[0, 0] == 11.0


class TestSimulatorImaging:
    def test__from_plane_and_grid__same_as_plane_image(self):

        psf = ag.Kernel2D.from_gaussian(
            shape_native=(7, 7), sigma=0.5, pixel_scales=0.05
        )

        grid = ag.Grid2D.uniform(shape_native=(20, 20), pixel_scales=0.05, sub_size=1)

        galaxy_0 = ag.Galaxy(redshift=0.5, light=ag.lp.EllipticalSersic(intensity=1.0))

        galaxy_1 = ag.Galaxy(redshift=1.0, light=ag.lp.EllipticalSersic(intensity=0.3))

        plane = ag.Plane(redshift=0.75, galaxies=[galaxy_0, galaxy_1])

        simulator = ag.SimulatorImaging(
            psf=psf,
            exposure_time=10000.0,
            background_sky_level=100.0,
            add_poisson_noise=False,
        )

        imaging = simulator.from_plane_and_grid(plane=plane, grid=grid)

        imaging_via_image = simulator.from_image(image=plane.image_from_grid(grid=grid))

        assert imaging.shape_native == (20, 20)
        assert imaging.image.native[0, 0] != imaging_via_image.image.native[0, 0]
        assert imaging.image.native[10, 10] == imaging_via_image.image.native[10, 10]
        assert (imaging.psf == imaging_via_image.psf).all()
        assert (imaging.noise_map == imaging_via_image.noise_map).all()

    def test__simulate_imaging_from_galaxy__source_galaxy__compare_to_imaging(self):

        galaxy_0 = ag.Galaxy(
            redshift=0.5,
            mass=ag.mp.EllipticalIsothermal(
                centre=(0.0, 0.0), einstein_radius=1.6, elliptical_comps=(0.17647, 0.0)
            ),
        )

        galaxy_1 = ag.Galaxy(
            redshift=1.0,
            light=ag.lp.EllipticalSersic(
                centre=(0.1, 0.1),
                elliptical_comps=(0.096225, -0.055555),
                intensity=0.3,
                effective_radius=1.0,
                sersic_index=2.5,
            ),
        )

        grid = ag.Grid2D.uniform(shape_native=(11, 11), pixel_scales=0.2, sub_size=1)

        psf = ag.Kernel2D.no_blur(pixel_scales=0.2)

        simulator = ag.SimulatorImaging(
            psf=psf,
            exposure_time=10000.0,
            background_sky_level=100.0,
            add_poisson_noise=True,
            noise_seed=1,
        )

        imaging = simulator.from_galaxies_and_grid(
            galaxies=[galaxy_0, galaxy_1], grid=grid
        )

        plane = ag.Plane(redshift=0.75, galaxies=[galaxy_0, galaxy_1])

        imaging_via_image = simulator.from_image(image=plane.image_from_grid(grid=grid))

        assert imaging.shape_native == (11, 11)
        assert (imaging.image == imaging_via_image.image).all()
        assert (imaging.psf == imaging_via_image.psf).all()
        assert imaging.noise_map == imaging_via_image.noise_map



from autogalaxy.profiles.mass_profiles.mass_profiles import psi_from
import numpy as np
from autoarray.structures import grids
from autogalaxy.profiles import mass_profiles as mp

from pyquad import quad_grid
from scipy.special import wofz
import typing
import copy

from autogalaxy.profiles.mass_profiles.mass_profiles import MassProfileMGE


class StellarProfile:

    pass


class EllipticalGaussian(mp.EllipticalMassProfile, StellarProfile):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        elliptical_comps: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        sigma: float = 0.01,
        mass_to_light_ratio: float = 1.0,
    ):
        """The elliptical Gaussian light profile.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        sigma : float
            The sigma value of the Gaussian.
        """

        super(EllipticalGaussian, self).__init__(
            centre=centre, elliptical_comps=elliptical_comps
        )
        super(mp.EllipticalMassProfile, self).__init__(
            centre=centre, elliptical_comps=elliptical_comps
        )
        self.mass_to_light_ratio = mass_to_light_ratio
        self.intensity = intensity
        self.sigma = sigma

        if self.axis_ratio > 0.9999:
            self.axis_ratio = 0.9999

    def zeta_from_grid(self, grid):
        q2 = self.axis_ratio ** 2.0
        ind_pos_y = grid[:, 0] >= 0
        shape_grid = np.shape(grid)
        output_grid = np.zeros((shape_grid[0]), dtype=np.complex128)
        scale_factor = self.axis_ratio / (self.sigma * np.sqrt(2.0 * (1.0 - q2)))

        xs_0 = grid[:, 1][ind_pos_y] * scale_factor
        ys_0 = grid[:, 0][ind_pos_y] * scale_factor
        xs_1 = grid[:, 1][~ind_pos_y] * scale_factor
        ys_1 = -grid[:, 0][~ind_pos_y] * scale_factor

        output_grid[ind_pos_y] = -1j * (
            wofz(xs_0 + 1j * ys_0)
            - np.exp(-(xs_0 ** 2.0) * (1.0 - q2) - ys_0 * ys_0 * (1.0 / q2 - 1.0))
            * wofz(self.axis_ratio * xs_0 + 1j * ys_0 / self.axis_ratio)
        )

        output_grid[~ind_pos_y] = np.conj(
            -1j
            * (
                wofz(xs_1 + 1j * ys_1)
                - np.exp(-(xs_1 ** 2.0) * (1.0 - q2) - ys_1 * ys_1 * (1.0 / q2 - 1.0))
                * wofz(self.axis_ratio * xs_1 + 1j * ys_1 / self.axis_ratio)
            )
        )

        return output_grid

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid2D
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        """

        deflections = (
            self.mass_to_light_ratio
            * self.intensity
            * self.sigma
            * np.sqrt((2 * np.pi) / (1.0 - self.axis_ratio ** 2.0))
            * self.zeta_from_grid(grid=grid)
        )

        return self.rotate_grid_from_profile(
            np.multiply(
                1.0, np.vstack((-1.0 * np.imag(deflections), np.real(deflections))).T
            )
        )

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def deflections_from_grid_via_integrator(self, grid):
        """
        Calculate the deflection angles at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid2D
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        Note: sigma is divided by sqrt(q) here.

        """

        def calculate_deflection_component(npow, index):
            deflection_grid = self.axis_ratio * grid[:, index]
            deflection_grid *= (
                self.intensity
                * self.mass_to_light_ratio
                * quad_grid(
                    self.deflection_func,
                    0.0,
                    1.0,
                    grid,
                    args=(npow, self.axis_ratio, self.sigma / np.sqrt(self.axis_ratio)),
                )[0]
            )

            return deflection_grid

        deflection_y = calculate_deflection_component(1.0, 0)
        deflection_x = calculate_deflection_component(0.0, 1)

        return self.rotate_grid_from_profile(
            np.multiply(1.0, np.vstack((deflection_y, deflection_x)).T)
        )

    @staticmethod
    def deflection_func(u, y, x, npow, axis_ratio, sigma):
        eta_u = np.sqrt(axis_ratio) * np.sqrt(
            (u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u))))
        )

        return np.exp(-0.5 * np.square(np.divide(eta_u, sigma))) / (
            (1 - (1 - axis_ratio ** 2) * u) ** (npow + 0.5)
        )

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def convergence_from_grid(self, grid):
        """Calculate the projected convergence at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid2D
            The grid of (y,x) arc-second coordinates the convergence is computed on.

        """
        return self.convergence_func(self.grid_to_eccentric_radii(grid))

    def convergence_func(self, grid_radius):
        return self.mass_to_light_ratio * self.image_from_grid_radii(grid_radius)

    def image_from_grid_radii(self, grid_radii):
        """Calculate the intensity of the Gaussian light profile on a grid of radial coordinates.

        Parameters
        ----------
        grid_radii : float
            The radial distance from the centre of the profile. for each coordinate on the grid.

        Note: sigma is divided by sqrt(q) here.
        """
        return np.multiply(
            self.intensity,
            np.exp(
                -0.5
                * np.square(
                    np.divide(grid_radii, self.sigma / np.sqrt(self.axis_ratio))
                )
            ),
        )

    def with_new_normalization(self, normalization):

        mass_profile = copy.copy(self)
        mass_profile.mass_to_light_ratio = normalization
        return mass_profile


# noinspection PyAbstractClass
class AbstractEllipticalSersic(
    mp.EllipticalMassProfile, MassProfileMGE, StellarProfile
):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        elliptical_comps: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        sersic_index: float = 0.6,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The Sersic mass profile, the mass profiles of the light profiles that are used to fit and subtract the lens \
        model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius : float
            The radius containing half the light of this profile.
        sersic_index : float
            Controls the concentration of the of the profile (lower value -> less concentrated, \
            higher value -> more concentrated).
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profiles
        """
        super(AbstractEllipticalSersic, self).__init__(
            centre=centre, elliptical_comps=elliptical_comps
        )
        super(mp.EllipticalMassProfile, self).__init__(
            centre=centre, elliptical_comps=elliptical_comps
        )
        super(MassProfileMGE, self).__init__()
        self.mass_to_light_ratio = mass_to_light_ratio
        self.intensity = intensity
        self.effective_radius = effective_radius
        self.sersic_index = sersic_index

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def convergence_from_grid(self, grid):
        """Calculate the projected convergence at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid2D
            The grid of (y,x) arc-second coordinates the convergence is computed on.

        """
        return self.convergence_func(self.grid_to_eccentric_radii(grid))

    def convergence_func(self, grid_radius):
        return self.mass_to_light_ratio * self.image_from_grid_radii(grid_radius)

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def convergence_from_grid_via_gaussians(self, grid):
        """Calculate the projected convergence at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid2D
            The grid of (y,x) arc-second coordinates the convergence is computed on.

        """

        eccentric_radii = self.grid_to_eccentric_radii(grid=grid)

        return self._convergence_from_grid_via_gaussians(grid_radii=eccentric_radii)

    @grids.grid_like_to_structure
    def potential_from_grid(self, grid):
        return np.zeros(shape=grid.shape[0])

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def deflections_from_grid(self, grid):
        return self._deflections_from_grid_via_gaussians(
            grid=grid, sigmas_factor=np.sqrt(self.axis_ratio)
        )

    @property
    def ellipticity_rescale(self):
        return 1.0 - ((1.0 - self.axis_ratio) / 2.0)

    def image_from_grid_radii(self, radius):
        """
        Returns the intensity of the profile at a given radius.

            Parameters
            ----------
            radius : float
                The distance from the centre of the profile.
        """
        return self.intensity * np.exp(
            -self.sersic_constant
            * (((radius / self.effective_radius) ** (1.0 / self.sersic_index)) - 1)
        )

    @property
    def sersic_constant(self):
        """A parameter derived from Sersic index which ensures that effective radius contains 50% of the profile's
        total integrated light.
        """
        return (
            (2 * self.sersic_index)
            - (1.0 / 3.0)
            + (4.0 / (405.0 * self.sersic_index))
            + (46.0 / (25515.0 * self.sersic_index ** 2))
            + (131.0 / (1148175.0 * self.sersic_index ** 3))
            - (2194697.0 / (30690717750.0 * self.sersic_index ** 4))
        )

    @property
    def elliptical_effective_radius(self):
        """The effective_radius of a Sersic light profile is defined as the circular effective radius. This is the \
        radius within which a circular aperture contains half the profiles's total integrated light. For elliptical \
        systems, this won't robustly capture the light profile's elliptical shape.

        The elliptical effective radius instead describes the major-axis radius of the ellipse containing \
        half the light, and may be more appropriate for highly flattened systems like disk galaxies."""
        return self.effective_radius / np.sqrt(self.axis_ratio)

    def decompose_convergence_into_gaussians(self):
        radii_min = self.effective_radius / 100.0
        radii_max = self.effective_radius * 20.0

        def sersic_2d(r):
            return (
                self.mass_to_light_ratio
                * self.intensity
                * np.exp(
                    -self.sersic_constant
                    * (((r / self.effective_radius) ** (1.0 / self.sersic_index)) - 1.0)
                )
            )

        return self._decompose_convergence_into_gaussians(
            func=sersic_2d, radii_min=radii_min, radii_max=radii_max
        )

    def with_new_normalization(self, normalization):

        mass_profile = copy.copy(self)
        mass_profile.mass_to_light_ratio = normalization
        return mass_profile


class EllipticalSersic(AbstractEllipticalSersic, MassProfileMGE):
    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def deflections_from_grid_via_integrator(self, grid):
        """
        Calculate the deflection angles at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid2D
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        """

        def calculate_deflection_component(npow, index):
            sersic_constant = self.sersic_constant

            deflection_grid = self.axis_ratio * grid[:, index]
            deflection_grid *= (
                self.intensity
                * self.mass_to_light_ratio
                * quad_grid(
                    self.deflection_func,
                    0.0,
                    1.0,
                    grid,
                    args=(
                        npow,
                        self.axis_ratio,
                        self.sersic_index,
                        self.effective_radius,
                        sersic_constant,
                    ),
                )[0]
            )

            return deflection_grid

        deflection_y = calculate_deflection_component(1.0, 0)
        deflection_x = calculate_deflection_component(0.0, 1)

        return self.rotate_grid_from_profile(
            np.multiply(1.0, np.vstack((deflection_y, deflection_x)).T)
        )

    @staticmethod
    def deflection_func(
        u, y, x, npow, axis_ratio, sersic_index, effective_radius, sersic_constant
    ):
        eta_u = np.sqrt(axis_ratio) * np.sqrt(
            (u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u))))
        )

        return np.exp(
            -sersic_constant
            * (((eta_u / effective_radius) ** (1.0 / sersic_index)) - 1)
        ) / ((1 - (1 - axis_ratio ** 2) * u) ** (npow + 0.5))


class SphericalSersic(EllipticalSersic):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        sersic_index: float = 0.6,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The Sersic mass profile, the mass profiles of the light profiles that are used to fit and subtract the lens
        model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second)
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : float
            Controls the concentration of the of the profile (lower value -> less concentrated, \
            higher value -> more concentrated).
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profile.
        """
        super(SphericalSersic, self).__init__(
            centre=centre,
            elliptical_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class EllipticalExponential(EllipticalSersic):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        elliptical_comps: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The EllipticalExponential mass profile, the mass profiles of the light profiles that are used to fit and
        subtract the lens model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profiles
        """
        super(EllipticalExponential, self).__init__(
            centre=centre,
            elliptical_comps=elliptical_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=1.0,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class SphericalExponential(EllipticalExponential):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The Exponential mass profile, the mass profiles of the light profiles that are used to fit and subtract the lens
        model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profiles.
        """
        super(SphericalExponential, self).__init__(
            centre=centre,
            elliptical_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class EllipticalDevVaucouleurs(EllipticalSersic):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        elliptical_comps: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The EllipticalDevVaucouleurs mass profile, the mass profiles of the light profiles that are used to fit and
        subtract the lens model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius : float
            The radius containing half the light of this profile.
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profile.
        """
        super(EllipticalDevVaucouleurs, self).__init__(
            centre=centre,
            elliptical_comps=elliptical_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=4.0,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class SphericalDevVaucouleurs(EllipticalDevVaucouleurs):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        mass_to_light_ratio: float = 1.0,
    ):
        """
        The DevVaucouleurs mass profile, the mass profiles of the light profiles that are used to fit and subtract the
        lens model_galaxy's light.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profiles.
        """
        super(SphericalDevVaucouleurs, self).__init__(
            centre=centre,
            elliptical_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
            mass_to_light_ratio=mass_to_light_ratio,
        )


class EllipticalSersicRadialGradient(AbstractEllipticalSersic):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        elliptical_comps: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        sersic_index: float = 0.6,
        mass_to_light_ratio: float = 1.0,
        mass_to_light_gradient: float = 0.0,
    ):
        """
        Setup a Sersic mass and light profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : float
            Controls the concentration of the of the profile (lower value -> less concentrated, \
            higher value -> more concentrated).
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profile.
        mass_to_light_gradient : float
            The mass-to-light radial gradient.
        """
        super(EllipticalSersicRadialGradient, self).__init__(
            centre=centre,
            elliptical_comps=elliptical_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            mass_to_light_ratio=mass_to_light_ratio,
        )
        self.mass_to_light_gradient = mass_to_light_gradient

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def convergence_from_grid(self, grid):
        """Calculate the projected convergence at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid2D
            The grid of (y,x) arc-second coordinates the convergence is computed on.

        """
        return self.convergence_func(self.grid_to_eccentric_radii(grid))

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def deflections_via_integrator_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of arc-second gridded coordinates.

        Parameters
        ----------
        grid : aa.Grid2D
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        """

        def calculate_deflection_component(npow, index):
            sersic_constant = self.sersic_constant

            deflection_grid = self.axis_ratio * grid[:, index]
            deflection_grid *= (
                self.intensity
                * self.mass_to_light_ratio
                * quad_grid(
                    self.deflection_func,
                    0.0,
                    1.0,
                    grid,
                    args=(
                        npow,
                        self.axis_ratio,
                        self.sersic_index,
                        self.effective_radius,
                        self.mass_to_light_gradient,
                        sersic_constant,
                    ),
                )[0]
            )
            return deflection_grid

        deflection_y = calculate_deflection_component(1.0, 0)
        deflection_x = calculate_deflection_component(0.0, 1)

        return self.rotate_grid_from_profile(
            np.multiply(1.0, np.vstack((deflection_y, deflection_x)).T)
        )

    def convergence_func(self, grid_radius):
        return (
            self.mass_to_light_ratio
            * (
                ((self.axis_ratio * grid_radius) / self.effective_radius)
                ** -self.mass_to_light_gradient
            )
            * self.image_from_grid_radii(grid_radius)
        )

    @staticmethod
    def deflection_func(
        u,
        y,
        x,
        npow,
        axis_ratio,
        sersic_index,
        effective_radius,
        mass_to_light_gradient,
        sersic_constant,
    ):
        eta_u = np.sqrt(axis_ratio) * np.sqrt(
            (u * ((x ** 2) + (y ** 2 / (1 - (1 - axis_ratio ** 2) * u))))
        )

        return (
            (((axis_ratio * eta_u) / effective_radius) ** -mass_to_light_gradient)
            * np.exp(
                -sersic_constant
                * (((eta_u / effective_radius) ** (1.0 / sersic_index)) - 1)
            )
            / ((1 - (1 - axis_ratio ** 2) * u) ** (npow + 0.5))
        )

    def decompose_convergence_into_gaussians(self):
        radii_min = self.effective_radius / 100.0
        radii_max = self.effective_radius * 20.0

        def sersic_radial_gradient_2D(r):
            return (
                self.mass_to_light_ratio
                * self.intensity
                * (
                    ((self.axis_ratio * r) / self.effective_radius)
                    ** -self.mass_to_light_gradient
                )
                * np.exp(
                    -self.sersic_constant
                    * (((r / self.effective_radius) ** (1.0 / self.sersic_index)) - 1.0)
                )
            )

        return self._decompose_convergence_into_gaussians(
            func=sersic_radial_gradient_2D, radii_min=radii_min, radii_max=radii_max
        )


class SphericalSersicRadialGradient(EllipticalSersicRadialGradient):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        sersic_index: float = 0.6,
        mass_to_light_ratio: float = 1.0,
        mass_to_light_gradient: float = 0.0,
    ):
        """
        Setup a Sersic mass and light profiles.

        Parameters
        ----------
        centre: (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        intensity : float
            Overall flux intensity normalisation in the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : float
            Controls the concentration of the of the profile (lower value -> less concentrated, \
            higher value -> more concentrated).
        mass_to_light_ratio : float
            The mass-to-light ratio of the light profile.
        mass_to_light_gradient : float
            The mass-to-light radial gradient.
        """
        super(SphericalSersicRadialGradient, self).__init__(
            centre=centre,
            elliptical_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            mass_to_light_ratio=mass_to_light_ratio,
            mass_to_light_gradient=mass_to_light_gradient,
        )


class EllipticalCoreSersic(EllipticalSersic):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        elliptical_comps: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        sersic_index: float = 4.0,
        radius_break: float = 0.01,
        intensity_break: float = 0.05,
        gamma: float = 0.25,
        alpha: float = 3.0,
        mass_to_light_ratio: float = 1.0,
    ):
        """ The elliptical cored-Sersic light profile.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : Int
            Controls the concentration of the of the profile (lower value -> less concentrated, \
            higher value -> more concentrated).
        radius_break : Float
            The break radius separating the inner power-law (with logarithmic slope gamma) and outer Sersic function.
        intensity_break : Float
            The intensity at the break radius.
        gamma : Float
            The logarithmic power-law slope of the inner core profiles
        alpha :
            Controls the sharpness of the transition between the inner core / outer Sersic profiles.
        """
        super(EllipticalCoreSersic, self).__init__(
            centre=centre,
            elliptical_comps=elliptical_comps,
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            mass_to_light_ratio=mass_to_light_ratio,
        )
        self.radius_break = radius_break
        self.intensity_break = intensity_break
        self.alpha = alpha
        self.gamma = gamma

    @property
    def intensity_prime(self):
        """Overall intensity normalisation in the rescaled Core-Sersic light profiles (electrons per second)"""
        return (
            self.intensity_break
            * (2.0 ** (-self.gamma / self.alpha))
            * np.exp(
                self.sersic_constant
                * (
                    ((2.0 ** (1.0 / self.alpha)) * self.radius_break)
                    / self.effective_radius
                )
                ** (1.0 / self.sersic_index)
            )
        )

    def image_from_grid_radii(self, grid_radii):
        """Calculate the intensity of the cored-Sersic light profile on a grid of radial coordinates.

        Parameters
        ----------
        grid_radii : float
            The radial distance from the centre of the profile. for each coordinate on the grid.
        """
        return np.multiply(
            np.multiply(
                self.intensity_prime,
                np.power(
                    np.add(
                        1,
                        np.power(np.divide(self.radius_break, grid_radii), self.alpha),
                    ),
                    (self.gamma / self.alpha),
                ),
            ),
            np.exp(
                np.multiply(
                    -self.sersic_constant,
                    (
                        np.power(
                            np.divide(
                                np.add(
                                    np.power(grid_radii, self.alpha),
                                    (self.radius_break ** self.alpha),
                                ),
                                (self.effective_radius ** self.alpha),
                            ),
                            (1.0 / (self.alpha * self.sersic_index)),
                        )
                    ),
                )
            ),
        )

    def decompose_convergence_into_gaussians(self):
        radii_min = self.effective_radius / 50.0
        radii_max = self.effective_radius * 20.0

        def core_sersic_2D(r):
            return (
                self.mass_to_light_ratio
                * self.intensity
                * (
                    self.intensity_prime
                    * (1.0 + (self.radius_break / r) ** self.alpha)
                    ** (self.gamma / self.alpha)
                    * np.exp(
                        -self.sersic_constant
                        * (
                            (r ** self.alpha + self.radius_break ** self.alpha)
                            / self.effective_radius ** self.alpha
                        )
                        ** (1.0 / (self.sersic_index * self.alpha))
                    )
                )
            )

        return self._decompose_convergence_into_gaussians(
            func=core_sersic_2D, radii_min=radii_min, radii_max=radii_max
        )


class SphericalCoreSersic(EllipticalCoreSersic):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        effective_radius: float = 0.6,
        sersic_index: float = 4.0,
        radius_break: float = 0.01,
        intensity_break: float = 0.05,
        gamma: float = 0.25,
        alpha: float = 3.0,
    ):
        """ The elliptical cored-Sersic light profile.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        effective_radius : float
            The circular radius containing half the light of this profile.
        sersic_index : Int
            Controls the concentration of the of the profile (lower value -> less concentrated, \
            higher value -> more concentrated).
        radius_break : Float
            The break radius separating the inner power-law (with logarithmic slope gamma) and outer Sersic function.
        intensity_break : Float
            The intensity at the break radius.
        gamma : Float
            The logarithmic power-law slope of the inner core profiles
        alpha :
            Controls the sharpness of the transition between the inner core / outer Sersic profiles.
        """
        super(SphericalCoreSersic, self).__init__(
            centre=centre,
            elliptical_comps=(0.0, 0.0),
            intensity=intensity,
            effective_radius=effective_radius,
            sersic_index=sersic_index,
            radius_break=radius_break,
            intensity_break=intensity_break,
            gamma=gamma,
            alpha=alpha,
        )
        self.radius_break = radius_break
        self.intensity_break = intensity_break
        self.alpha = alpha
        self.gamma = gamma


class EllipticalChameleon(mp.EllipticalMassProfile, StellarProfile):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        elliptical_comps: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        core_radius_0: float = 0.01,
        core_radius_1: float = 0.02,
        mass_to_light_ratio: float = 1.0,
    ):
        """ The elliptical Chamelon mass profile.

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        core_radius_0 : the core size of the first elliptical cored Isothermal profile.
        core_radius_1 : core_radius_0 + core_radius_1 is the core size of the second elliptical cored Isothermal profile.
            We use core_radius_1 here is to avoid negative values.

        Profile form:
            mass_to_light_ratio * intensity *\
                (1.0 / Sqrt(x^2 + (y/q)^2 + core_radius_0^2) - 1.0 / Sqrt(x^2 + (y/q)^2 + (core_radius_0 + core_radius_1)**2.0))
        """

        super(EllipticalChameleon, self).__init__(
            centre=centre, elliptical_comps=elliptical_comps
        )
        super(mp.EllipticalMassProfile, self).__init__(
            centre=centre, elliptical_comps=elliptical_comps
        )
        self.mass_to_light_ratio = mass_to_light_ratio
        self.intensity = intensity
        self.core_radius_0 = core_radius_0
        self.core_radius_1 = core_radius_1
        if self.axis_ratio > 0.99999:
            self.axis_ratio = 0.99999

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def deflections_from_grid(self, grid):
        """
        Calculate the deflection angles at a given set of arc-second gridded coordinates.
        Following Eq. (15) and (16), but the parameters are slightly different.

        Parameters
        ----------
        grid : aa.Grid2D
            The grid of (y,x) arc-second coordinates the deflection angles are computed on.

        """

        factor = (
            2.0
            * self.mass_to_light_ratio
            * self.intensity
            / (1 + self.axis_ratio)
            * self.axis_ratio
            / np.sqrt(1.0 - self.axis_ratio ** 2.0)
        )

        core_radius_0 = np.sqrt(
            (4.0 * self.core_radius_0 ** 2.0) / (1.0 + self.axis_ratio) ** 2
        )
        core_radius_1 = np.sqrt(
            (4.0 * self.core_radius_1 ** 2.0) / (1.0 + self.axis_ratio) ** 2
        )

        psi0 = psi_from(
            grid=grid, axis_ratio=self.axis_ratio, core_radius=core_radius_0
        )
        psi1 = psi_from(
            grid=grid, axis_ratio=self.axis_ratio, core_radius=core_radius_1
        )

        deflection_y0 = np.arctanh(
            np.divide(
                np.multiply(np.sqrt(1.0 - self.axis_ratio ** 2.0), grid[:, 0]),
                np.add(psi0, self.axis_ratio ** 2.0 * core_radius_0),
            )
        )

        deflection_x0 = np.arctan(
            np.divide(
                np.multiply(np.sqrt(1.0 - self.axis_ratio ** 2.0), grid[:, 1]),
                np.add(psi0, core_radius_0),
            )
        )

        deflection_y1 = np.arctanh(
            np.divide(
                np.multiply(np.sqrt(1.0 - self.axis_ratio ** 2.0), grid[:, 0]),
                np.add(psi1, self.axis_ratio ** 2.0 * core_radius_1),
            )
        )

        deflection_x1 = np.arctan(
            np.divide(
                np.multiply(np.sqrt(1.0 - self.axis_ratio ** 2.0), grid[:, 1]),
                np.add(psi1, core_radius_1),
            )
        )

        deflection_y = np.subtract(deflection_y0, deflection_y1)
        deflection_x = np.subtract(deflection_x0, deflection_x1)

        return self.rotate_grid_from_profile(
            np.multiply(factor, np.vstack((deflection_y, deflection_x)).T)
        )

    @grids.grid_like_to_structure
    @grids.transform
    @grids.relocate_to_radial_minimum
    def convergence_from_grid(self, grid):
        """Calculate the projected convergence at a given set of arc-second gridded coordinates.
        Parameters
        ----------
        grid : aa.Grid2D
            The grid of (y,x) arc-second coordinates the convergence is computed on.
        """
        return self.convergence_func(self.grid_to_elliptical_radii(grid))

    def convergence_func(self, grid_radius):
        return self.mass_to_light_ratio * self.image_from_grid_radii(grid_radius)

    @grids.grid_like_to_structure
    def potential_from_grid(self, grid):
        return np.zeros(shape=grid.shape[0])

    def image_from_grid_radii(self, grid_radii):
        """Calculate the intensity of the Chamelon light profile on a grid of radial coordinates.

        Parameters
        ----------
        grid_radii : float
            The radial distance from the centre of the profile. for each coordinate on the grid.
        """

        axis_ratio_factor = (1.0 + self.axis_ratio) ** 2.0

        return np.multiply(
            self.intensity / (1 + self.axis_ratio),
            np.add(
                np.divide(
                    1.0,
                    np.sqrt(
                        np.add(
                            np.square(grid_radii),
                            (4.0 * self.core_radius_0 ** 2.0) / axis_ratio_factor,
                        )
                    ),
                ),
                -np.divide(
                    1.0,
                    np.sqrt(
                        np.add(
                            np.square(grid_radii),
                            (4.0 * self.core_radius_1 ** 2.0) / axis_ratio_factor,
                        )
                    ),
                ),
            ),
        )

    def with_new_normalization(self, normalization):

        mass_profile = copy.copy(self)
        mass_profile.mass_to_light_ratio = normalization
        return mass_profile


class SphericalChameleon(EllipticalChameleon):
    def __init__(
        self,
        centre: typing.Tuple[float, float] = (0.0, 0.0),
        intensity: float = 0.1,
        core_radius_0: float = 0.01,
        core_radius_1: float = 0.02,
        mass_to_light_ratio: float = 1.0,
    ):
        """ The spherica; Chameleon mass profile.

        Profile form:
            mass_to_light_ratio * intensity *\
                (1.0 / Sqrt(x^2 + (y/q)^2 + core_radius_0^2) - 1.0 / Sqrt(x^2 + (y/q)^2 + (core_radius_0 + core_radius_1)**2.0))

        Parameters
        ----------
        centre : (float, float)
            The (y,x) arc-second coordinates of the profile centre.
        elliptical_comps : (float, float)
            The first and second ellipticity components of the elliptical coordinate system, where
            fac = (1 - axis_ratio) / (1 + axis_ratio), ellip_y = fac * sin(2*phi) and ellip_x = fac * cos(2*phi).
        intensity : float
            Overall intensity normalisation of the light profiles (electrons per second).
        core_radius_0 : the core size of the first elliptical cored Isothermal profile.
        core_radius_1 : core_radius_0 + core_radius_1 is the core size of the second elliptical cored Isothermal profile.
            We use core_radius_1 here is to avoid negative values.
       """

        super(SphericalChameleon, self).__init__(
            centre=centre,
            elliptical_comps=(0.0, 0.0),
            intensity=intensity,
            core_radius_0=core_radius_0,
            core_radius_1=core_radius_1,
            mass_to_light_ratio=mass_to_light_ratio,
        )
